"""Compose module for programmatic music creation.

This module provides domain objects for building music programmatically,
generating AST nodes directly without parsing text.

Examples:
    >>> from aldakit.compose import note, rest, chord, seq, part, tempo
    >>> from aldakit import Score
    >>>
    >>> score = Score.from_elements(
    ...     part("piano"),
    ...     tempo(120),
    ...     note("c", duration=4),
    ...     note("d"),
    ...     note("e"),
    ... )
    >>> score.play()
"""

from .attributes import (
    Dynamic,
    OctaveDown,
    OctaveSet,
    OctaveUp,
    Panning,
    Quant,
    Tempo,
    Volume,
    f,
    ff,
    mf,
    mp,
    octave,
    octave_down,
    octave_up,
    p,
    panning,
    pp,
    quant,
    tempo,
    vol,
    volume,
)
from .base import ComposeElement
from .chords import (
    CHORD_INTERVALS,
    add9,
    arpeggiate,
    aug,
    aug7,
    build_chord,
    dim,
    dim7,
    dom7,
    dom9,
    half_dim7,
)
from .chords import invert as invert_chord
from .chords import (
    list_chord_types,
    maj6,
    maj7,
    maj9,
    major,
    min6,
    min7,
    min9,
    min_maj7,
    minor,
    power,
    sus2,
    sus4,
    voicing,
)
from .core import (
    AtMarker,
    Chord,
    Cram,
    Marker,
    Note,
    Repeat,
    Rest,
    Seq,
    Variable,
    VariableRef,
    Voice,
    VoiceGroup,
    at_marker,
    chord,
    cram,
    marker,
    note,
    rest,
    seq,
    var,
    var_ref,
    voice,
    voice_group,
)
from .generate import (  # Random selection; Random walks; Rhythmic generators; Markov chains; L-Systems; Cellular automata; Shift registers
    MarkovChain,
    cellular_automaton,
    drunk_walk,
    euclidean,
    learn_markov,
    lsystem,
    markov_chain,
    probability_seq,
    random_choice,
    random_note,
    random_walk,
    rest_probability,
    shift_register,
    turing_machine,
    weighted_choice,
)
from .part import Part, part
from .scales import (
    SCALE_INTERVALS,
    interval_name,
    list_scales,
    mode,
    parallel_major,
    parallel_minor,
    relative_major,
    relative_minor,
    scale,
    scale_degree,
    scale_notes,
    transpose_scale,
)
from .transform import (  # Pitch transformers; Structural transformers; Helpers
    augment,
    concat,
    diminish,
    fragment,
    identity,
    interleave,
    invert,
    loop,
    pipe,
    retrograde_inversion,
    reverse,
    rotate,
    shuffle,
    split,
    take_every,
    transpose,
)

__all__ = [
    # Base
    "ComposeElement",
    # Core classes
    "Note",
    "Rest",
    "Chord",
    "Seq",
    "Repeat",
    "Cram",
    "Voice",
    "VoiceGroup",
    "Variable",
    "VariableRef",
    "Marker",
    "AtMarker",
    # Core factory functions
    "note",
    "rest",
    "chord",
    "seq",
    "cram",
    "voice",
    "voice_group",
    "var",
    "var_ref",
    "marker",
    "at_marker",
    # Part
    "Part",
    "part",
    # Attribute classes
    "Tempo",
    "Volume",
    "Quant",
    "Panning",
    "OctaveSet",
    "OctaveUp",
    "OctaveDown",
    "Dynamic",
    # Attribute factory functions
    "tempo",
    "volume",
    "vol",
    "quant",
    "panning",
    "octave",
    "octave_up",
    "octave_down",
    # Dynamics
    "pp",
    "p",
    "mp",
    "mf",
    "f",
    "ff",
    # Pitch transformers
    "transpose",
    "invert",
    "reverse",
    "shuffle",
    "retrograde_inversion",
    # Structural transformers
    "augment",
    "diminish",
    "fragment",
    "loop",
    "interleave",
    "rotate",
    "take_every",
    "split",
    "concat",
    # Helpers
    "pipe",
    "identity",
    # Random selection
    "random_note",
    "random_choice",
    "weighted_choice",
    # Random walks
    "random_walk",
    "drunk_walk",
    # Rhythmic generators
    "euclidean",
    "probability_seq",
    "rest_probability",
    # Markov chains
    "MarkovChain",
    "markov_chain",
    "learn_markov",
    # L-Systems
    "lsystem",
    # Cellular automata
    "cellular_automaton",
    # Shift registers
    "shift_register",
    "turing_machine",
    # Scales
    "scale",
    "scale_notes",
    "scale_degree",
    "mode",
    "relative_minor",
    "relative_major",
    "parallel_minor",
    "parallel_major",
    "transpose_scale",
    "interval_name",
    "list_scales",
    "SCALE_INTERVALS",
    # Chords
    "build_chord",
    "major",
    "minor",
    "dim",
    "aug",
    "sus2",
    "sus4",
    "maj7",
    "min7",
    "dom7",
    "dim7",
    "half_dim7",
    "min_maj7",
    "aug7",
    "maj6",
    "min6",
    "dom9",
    "maj9",
    "min9",
    "add9",
    "power",
    "arpeggiate",
    "invert_chord",
    "voicing",
    "list_chord_types",
    "CHORD_INTERVALS",
]



"""
Margin implementations for a :class:`~prompt_toolkit.layout.containers.Window`.
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Callable

from prompt_toolkit.filters import FilterOrBool, to_filter
from prompt_toolkit.formatted_text import (StyleAndTextTuples,
                                           fragment_list_to_text,
                                           to_formatted_text)
from prompt_toolkit.utils import get_cwidth

from .controls import UIContent

if TYPE_CHECKING:
    from .containers import WindowRenderInfo

__all__ = [
    "Margin",
    "NumberedMargin",
    "ScrollbarMargin",
    "ConditionalMargin",
    "PromptMargin",
]


class Margin(metaclass=ABCMeta):
    """
    Base interface for a margin.
    """

    @abstractmethod
    def get_width(self, get_ui_content: Callable[[], UIContent]) -> int:
        """
        Return the width that this margin is going to consume.

        :param get_ui_content: Callable that asks the user control to create
            a :class:`.UIContent` instance. This can be used for instance to
            obtain the number of lines.
        """
        return 0

    @abstractmethod
    def create_margin(
        self, window_render_info: WindowRenderInfo, width: int, height: int
    ) -> StyleAndTextTuples:
        """
        Creates a margin.
        This should return a list of (style_str, text) tuples.

        :param window_render_info:
            :class:`~prompt_toolkit.layout.containers.WindowRenderInfo`
            instance, generated after rendering and copying the visible part of
            the :class:`~prompt_toolkit.layout.controls.UIControl` into the
            :class:`~prompt_toolkit.layout.containers.Window`.
        :param width: The width that's available for this margin. (As reported
            by :meth:`.get_width`.)
        :param height: The height that's available for this margin. (The height
            of the :class:`~prompt_toolkit.layout.containers.Window`.)
        """
        return []


class NumberedMargin(Margin):
    """
    Margin that displays the line numbers.

    :param relative: Number relative to the cursor position. Similar to the Vi
                     'relativenumber' option.
    :param display_tildes: Display tildes after the end of the document, just
        like Vi does.
    """

    def __init__(
        self, relative: FilterOrBool = False, display_tildes: FilterOrBool = False
    ) -> None:
        self.relative = to_filter(relative)
        self.display_tildes = to_filter(display_tildes)

    def get_width(self, get_ui_content: Callable[[], UIContent]) -> int:
        line_count = get_ui_content().line_count
        return max(3, len(f"{line_count}") + 1)

    def create_margin(
        self, window_render_info: WindowRenderInfo, width: int, height: int
    ) -> StyleAndTextTuples:
        relative = self.relative()

        style = "class:line-number"
        style_current = "class:line-number.current"

        # Get current line number.
        current_lineno = window_render_info.ui_content.cursor_position.y

        # Construct margin.
        result: StyleAndTextTuples = []
        last_lineno = None

        for y, lineno in enumerate(window_render_info.displayed_lines):
            # Only display line number if this line is not a continuation of the previous line.
            if lineno != last_lineno:
                if lineno is None:
                    pass
                elif lineno == current_lineno:
                    # Current line.
                    if relative:
                        # Left align current number in relative mode.
                        result.append((style_current, "%i" % (lineno + 1)))
                    else:
                        result.append(
                            (style_current, ("%i " % (lineno + 1)).rjust(width))
                        )
                else:
                    # Other lines.
                    if relative:
                        lineno = abs(lineno - current_lineno) - 1

                    result.append((style, ("%i " % (lineno + 1)).rjust(width)))

            last_lineno = lineno
            result.append(("", "\n"))

        # Fill with tildes.
        if self.display_tildes():
            while y < window_render_info.window_height:
                result.append(("class:tilde", "~\n"))
                y += 1

        return result


class ConditionalMargin(Margin):
    """
    Wrapper around other :class:`.Margin` classes to show/hide them.
    """

    def __init__(self, margin: Margin, filter: FilterOrBool) -> None:
        self.margin = margin
        self.filter = to_filter(filter)

    def get_width(self, get_ui_content: Callable[[], UIContent]) -> int:
        if self.filter():
            return self.margin.get_width(get_ui_content)
        else:
            return 0

    def create_margin(
        self, window_render_info: WindowRenderInfo, width: int, height: int
    ) -> StyleAndTextTuples:
        if width and self.filter():
            return self.margin.create_margin(window_render_info, width, height)
        else:
            return []


class ScrollbarMargin(Margin):
    """
    Margin displaying a scrollbar.

    :param display_arrows: Display scroll up/down arrows.
    """

    def __init__(
        self,
        display_arrows: FilterOrBool = False,
        up_arrow_symbol: str = "^",
        down_arrow_symbol: str = "v",
    ) -> None:
        self.display_arrows = to_filter(display_arrows)
        self.up_arrow_symbol = up_arrow_symbol
        self.down_arrow_symbol = down_arrow_symbol

    def get_width(self, get_ui_content: Callable[[], UIContent]) -> int:
        return 1

    def create_margin(
        self, window_render_info: WindowRenderInfo, width: int, height: int
    ) -> StyleAndTextTuples:
        content_height = window_render_info.content_height
        window_height = window_render_info.window_height
        display_arrows = self.display_arrows()

        if display_arrows:
            window_height -= 2

        try:
            fraction_visible = len(window_render_info.displayed_lines) / float(
                content_height
            )
            fraction_above = window_render_info.vertical_scroll / float(content_height)

            scrollbar_height = int(
                min(window_height, max(1, window_height * fraction_visible))
            )
            scrollbar_top = int(window_height * fraction_above)
        except ZeroDivisionError:
            return []
        else:

            def is_scroll_button(row: int) -> bool:
                "True if we should display a button on this row."
                return scrollbar_top <= row <= scrollbar_top + scrollbar_height

            # Up arrow.
            result: StyleAndTextTuples = []
            if display_arrows:
                result.extend(
                    [
                        ("class:scrollbar.arrow", self.up_arrow_symbol),
                        ("class:scrollbar", "\n"),
                    ]
                )

            # Scrollbar body.
            scrollbar_background = "class:scrollbar.background"
            scrollbar_background_start = "class:scrollbar.background,scrollbar.start"
            scrollbar_button = "class:scrollbar.button"
            scrollbar_button_end = "class:scrollbar.button,scrollbar.end"

            for i in range(window_height):
                if is_scroll_button(i):
                    if not is_scroll_button(i + 1):
                        # Give the last cell a different style, because we
                        # want to underline this.
                        result.append((scrollbar_button_end, " "))
                    else:
                        result.append((scrollbar_button, " "))
                else:
                    if is_scroll_button(i + 1):
                        result.append((scrollbar_background_start, " "))
                    else:
                        result.append((scrollbar_background, " "))
                result.append(("", "\n"))

            # Down arrow
            if display_arrows:
                result.append(("class:scrollbar.arrow", self.down_arrow_symbol))

            return result


class PromptMargin(Margin):
    """
    [Deprecated]

    Create margin that displays a prompt.
    This can display one prompt at the first line, and a continuation prompt
    (e.g, just dots) on all the following lines.

    This `PromptMargin` implementation has been largely superseded in favor of
    the `get_line_prefix` attribute of `Window`. The reason is that a margin is
    always a fixed width, while `get_line_prefix` can return a variable width
    prefix in front of every line, making it more powerful, especially for line
    continuations.

    :param get_prompt: Callable returns formatted text or a list of
        `(style_str, type)` tuples to be shown as the prompt at the first line.
    :param get_continuation: Callable that takes three inputs. The width (int),
        line_number (int), and is_soft_wrap (bool). It should return formatted
        text or a list of `(style_str, type)` tuples for the next lines of the
        input.
    """

    def __init__(
        self,
        get_prompt: Callable[[], StyleAndTextTuples],
        get_continuation: None
        | (Callable[[int, int, bool], StyleAndTextTuples]) = None,
    ) -> None:
        self.get_prompt = get_prompt
        self.get_continuation = get_continuation

    def get_width(self, get_ui_content: Callable[[], UIContent]) -> int:
        "Width to report to the `Window`."
        # Take the width from the first line.
        text = fragment_list_to_text(self.get_prompt())
        return get_cwidth(text)

    def create_margin(
        self, window_render_info: WindowRenderInfo, width: int, height: int
    ) -> StyleAndTextTuples:
        get_continuation = self.get_continuation
        result: StyleAndTextTuples = []

        # First line.
        result.extend(to_formatted_text(self.get_prompt()))

        # Next lines.
        if get_continuation:
            last_y = None

            for y in window_render_info.displayed_lines[1:]:
                result.append(("", "\n"))
                result.extend(
                    to_formatted_text(get_continuation(width, y, y == last_y))
                )
                last_y = y

        return result



#!/usr/bin/env python
import pylibremidi as lm

api = lm.midi2_default_api()
print(api)
observer_config = lm.ObserverConfiguration()
observer = lm.Observer(observer_config, api)

pi = observer.get_input_ports()
if len(pi) == 0:
    print("No input available")
    exit(1)
else:
    print(f"Found input: {pi[0].display_name}")

po = observer.get_output_ports()
if len(po) == 0:
    print("No output available")
    exit(1)
else:
    print(f"Found output: {po[0].display_name}")

out_config = lm.OutputConfiguration()
midi_out = lm.MidiOut(out_config, api)
err = midi_out.open_port(po[0])
if err:
    print("Error when opening output port: ", err)
    exit(1)

in_config = lm.UmpInputConfiguration()
in_config.on_message = lambda msg: print(f"{msg}")

midi_in = lm.MidiIn(in_config, api)
err = midi_in.open_port(pi[0])
if err:
    print("Error when opening input port: ", err)
    exit(1)

while True:
    midi_in.poll()



"""aldakit: a pythonic alda music programming language implementation."""

from .api import list_ports, play, play_file, save, save_file
from .ast_nodes import (
    ASTNode,
    ASTVisitor,
    AtMarkerNode,  # Phase 2 nodes
    BarlineNode,
    BracketedSequenceNode,
    ChordNode,
    CramNode,
    DurationNode,
    EventSequenceNode,
    LispListNode,
    LispNumberNode,
    LispStringNode,
    LispSymbolNode,
    MarkerNode,
    NoteLengthMsNode,
    NoteLengthNode,
    NoteLengthSecondsNode,
    NoteNode,
    OctaveDownNode,
    OctaveSetNode,
    OctaveUpNode,
    OnRepetitionsNode,
    PartDeclarationNode,
    PartNode,
    RepeatNode,
    RepetitionRange,
    RestNode,
    RootNode,
    VariableDefinitionNode,
    VariableReferenceNode,
    VoiceGroupNode,
    VoiceNode,
)
from .errors import AldaParseError, AldaScanError, AldaSyntaxError
from .midi import (
    LibremidiBackend,
    MidiBackend,
    MidiGenerator,
    MidiNote,
    MidiSequence,
    generate_midi,
)
from .midi.transcriber import list_input_ports, transcribe
from .parser import Parser, parse
from .scanner import Scanner
from .score import Score
from .tokens import SourcePosition, Token, TokenType

__version__ = "0.1.7"


__all__ = [
    # High-level API
    "Score",
    "play",
    "play_file",
    "save",
    "save_file",
    "list_ports",
    "transcribe",
    "list_input_ports",
    # Convenience function
    "parse",
    # Core classes
    "Token",
    "TokenType",
    "SourcePosition",
    "Scanner",
    "Parser",
    # AST nodes - Core
    "ASTNode",
    "ASTVisitor",
    "RootNode",
    "PartNode",
    "PartDeclarationNode",
    "EventSequenceNode",
    "NoteNode",
    "RestNode",
    "ChordNode",
    "DurationNode",
    "NoteLengthNode",
    "NoteLengthMsNode",
    "NoteLengthSecondsNode",
    "BarlineNode",
    "OctaveSetNode",
    "OctaveUpNode",
    "OctaveDownNode",
    "LispListNode",
    "LispSymbolNode",
    "LispNumberNode",
    "LispStringNode",
    # AST nodes - Phase 2
    "VariableDefinitionNode",
    "VariableReferenceNode",
    "MarkerNode",
    "AtMarkerNode",
    "VoiceNode",
    "VoiceGroupNode",
    "CramNode",
    "RepeatNode",
    "OnRepetitionsNode",
    "RepetitionRange",
    "BracketedSequenceNode",
    # Errors
    "AldaParseError",
    "AldaScanError",
    "AldaSyntaxError",
    # MIDI
    "MidiSequence",
    "MidiNote",
    "MidiGenerator",
    "generate_midi",
    "MidiBackend",
    "LibremidiBackend",
]



"""Tests for the compose module."""

import pytest

from aldakit import Score
from aldakit.compose import (
    # Core
    Note,
    Seq,
    Repeat,
    note,
    rest,
    chord,
    seq,
    # Part
    part,
    # Attributes
    tempo,
    volume,
    vol,
    quant,
    panning,
    octave,
    octave_up,
    octave_down,
    pp,
    p,
    mp,
    mf,
    f,
    ff,
)
from aldakit.ast_nodes import (
    NoteNode,
    RestNode,
    ChordNode,
    EventSequenceNode,
    PartDeclarationNode,
    PartNode,
    LispListNode,
    OctaveSetNode,
    OctaveUpNode,
    OctaveDownNode,
    RepeatNode,
)


class TestNote:
    """Test Note class and note() factory."""

    def test_basic_note(self):
        n = note("c")
        assert n.pitch == "c"
        assert n.duration is None
        assert n.octave is None
        assert n.accidental is None

    def test_note_with_duration(self):
        n = note("c", duration=4)
        assert n.duration == 4

    def test_note_with_octave(self):
        n = note("c", octave=5)
        assert n.octave == 5

    def test_note_with_accidental(self):
        n = note("c", accidental="+")
        assert n.accidental == "+"

    def test_note_with_dots(self):
        n = note("c", duration=4, dots=1)
        assert n.dots == 1

    def test_note_with_ms(self):
        n = note("c", ms=500)
        assert n.ms == 500

    def test_note_with_seconds(self):
        n = note("c", seconds=2)
        assert n.seconds == 2

    def test_note_slurred(self):
        n = note("c", slurred=True)
        assert n.slurred is True

    def test_note_invalid_pitch(self):
        with pytest.raises(ValueError):
            note("x")

    def test_note_to_ast(self):
        n = note("c", duration=4)
        ast = n.to_ast()
        assert isinstance(ast, NoteNode)
        assert ast.letter == "c"
        assert ast.duration is not None

    def test_note_to_ast_with_accidentals(self):
        n = note("c", accidental="++")
        ast = n.to_ast()
        assert ast.accidentals == ["+", "+"]

    def test_note_to_alda(self):
        assert note("c").to_alda() == "c"
        assert note("c", duration=4).to_alda() == "c4"
        assert note("c", accidental="+").to_alda() == "c+"
        assert note("c", duration=4, dots=1).to_alda() == "c4."
        assert note("c", ms=500).to_alda() == "c500ms"
        assert note("c", seconds=2).to_alda() == "c2s"
        assert note("c", slurred=True).to_alda() == "c~"

    def test_note_midi_pitch(self):
        assert note("c").midi_pitch == 60  # C4
        assert note("c", octave=5).midi_pitch == 72  # C5
        assert note("c", accidental="+").midi_pitch == 61  # C#4
        assert note("c", accidental="-").midi_pitch == 59  # Cb4 (B3)

    def test_note_sharpen(self):
        n = note("c")
        sharp = n.sharpen()
        assert sharp.accidental == "+"
        assert n.accidental is None  # Original unchanged

    def test_note_flatten(self):
        n = note("c")
        flat = n.flatten()
        assert flat.accidental == "-"

    def test_note_transpose(self):
        n = note("c", octave=4)
        up = n.transpose(semitones=2)
        assert up.pitch == "d"
        assert up.octave == 4

    def test_note_transpose_with_accidental(self):
        n = note("c", octave=4)
        up = n.transpose(semitones=1)
        assert up.pitch == "c"
        assert up.accidental == "+"

    def test_note_with_duration_method(self):
        n = note("c")
        n2 = n.with_duration(8)
        assert n2.duration == 8
        assert n.duration is None  # Original unchanged

    def test_note_with_octave_method(self):
        n = note("c")
        n2 = n.with_octave(5)
        assert n2.octave == 5

    def test_note_slur_method(self):
        n = note("c")
        slurred = n.slur()
        assert slurred.slurred is True


class TestRest:
    """Test Rest class and rest() factory."""

    def test_basic_rest(self):
        r = rest()
        assert r.duration is None

    def test_rest_with_duration(self):
        r = rest(duration=2)
        assert r.duration == 2

    def test_rest_with_ms(self):
        r = rest(ms=1000)
        assert r.ms == 1000

    def test_rest_to_ast(self):
        r = rest(duration=4)
        ast = r.to_ast()
        assert isinstance(ast, RestNode)

    def test_rest_to_alda(self):
        assert rest().to_alda() == "r"
        assert rest(duration=2).to_alda() == "r2"
        assert rest(ms=1000).to_alda() == "r1000ms"


class TestChord:
    """Test Chord class and chord() factory."""

    def test_chord_from_strings(self):
        c = chord("c", "e", "g")
        assert len(c.notes) == 3
        assert all(isinstance(n, Note) for n in c.notes)

    def test_chord_from_notes(self):
        c = chord(note("c"), note("e"), note("g"))
        assert len(c.notes) == 3

    def test_chord_with_duration(self):
        c = chord("c", "e", "g", duration=1)
        assert c.duration == 1

    def test_chord_to_ast(self):
        c = chord("c", "e", "g")
        ast = c.to_ast()
        assert isinstance(ast, ChordNode)
        assert len(ast.notes) == 3

    def test_chord_to_alda(self):
        assert chord("c", "e", "g").to_alda() == "c/e/g"
        assert chord("c", "e", "g", duration=1).to_alda() == "c1/e/g"


class TestSeq:
    """Test Seq class and seq() factory."""

    def test_basic_seq(self):
        s = seq(note("c"), note("d"), note("e"))
        assert len(s.elements) == 3

    def test_seq_to_ast(self):
        s = seq(note("c"), note("d"))
        ast = s.to_ast()
        assert isinstance(ast, EventSequenceNode)
        assert len(ast.events) == 2

    def test_seq_to_alda(self):
        s = seq(note("c"), note("d"), note("e"))
        assert s.to_alda() == "c d e"

    def test_seq_multiply(self):
        s = seq(note("c"), note("d"))
        repeated = s * 4
        assert isinstance(repeated, Repeat)
        assert repeated.times == 4

    def test_seq_rmultiply(self):
        s = seq(note("c"), note("d"))
        repeated = 4 * s
        assert isinstance(repeated, Repeat)
        assert repeated.times == 4

    def test_seq_add(self):
        s1 = seq(note("c"), note("d"))
        s2 = seq(note("e"), note("f"))
        combined = s1 + s2
        assert len(combined.elements) == 4

    def test_seq_from_alda(self):
        s = Seq.from_alda("c d e f g")
        ast = s.to_ast()
        assert isinstance(ast, EventSequenceNode)


class TestRepeat:
    """Test Repeat class."""

    def test_repeat_note(self):
        r = note("c") * 4
        assert r.times == 4

    def test_repeat_to_ast(self):
        r = note("c") * 4
        ast = r.to_ast()
        assert isinstance(ast, RepeatNode)

    def test_repeat_to_alda(self):
        r = note("c") * 4
        assert r.to_alda() == "c*4"

    def test_repeat_seq_to_alda(self):
        s = seq(note("c"), note("d"))
        r = s * 4
        assert r.to_alda() == "[c d]*4"


class TestPart:
    """Test Part class and part() factory."""

    def test_basic_part(self):
        p = part("piano")
        assert p.instruments == ("piano",)
        assert p.alias is None

    def test_part_with_alias(self):
        p = part("violin", alias="v1")
        assert p.alias == "v1"

    def test_multi_instrument_part(self):
        p = part("violin", "viola", "cello", alias="strings")
        assert p.instruments == ("violin", "viola", "cello")

    def test_part_empty_raises(self):
        with pytest.raises(ValueError):
            part()

    def test_part_to_ast(self):
        p = part("piano")
        ast = p.to_ast()
        assert isinstance(ast, PartDeclarationNode)
        assert ast.names == ["piano"]

    def test_part_to_alda(self):
        assert part("piano").to_alda() == "piano:"
        assert part("violin", alias="v1").to_alda() == 'violin "v1":'
        assert part("violin", "viola").to_alda() == "violin/viola:"


class TestTempo:
    """Test tempo attribute."""

    def test_tempo(self):
        t = tempo(120)
        assert t.bpm == 120
        assert t.global_ is False

    def test_tempo_global(self):
        t = tempo(120, global_=True)
        assert t.global_ is True

    def test_tempo_to_ast(self):
        t = tempo(120)
        ast = t.to_ast()
        assert isinstance(ast, LispListNode)

    def test_tempo_to_alda(self):
        assert tempo(120).to_alda() == "(tempo 120)"
        assert tempo(120, global_=True).to_alda() == "(tempo! 120)"


class TestVolume:
    """Test volume attribute."""

    def test_volume(self):
        v = volume(80)
        assert v.level == 80

    def test_vol_alias(self):
        v = vol(80)
        assert v.level == 80

    def test_volume_to_alda(self):
        assert volume(80).to_alda() == "(volume 80)"


class TestOctave:
    """Test octave attributes."""

    def test_octave_set(self):
        o = octave(5)
        assert o.value == 5

    def test_octave_set_to_ast(self):
        o = octave(5)
        ast = o.to_ast()
        assert isinstance(ast, OctaveSetNode)
        assert ast.octave == 5

    def test_octave_set_to_alda(self):
        assert octave(5).to_alda() == "o5"

    def test_octave_up(self):
        o = octave_up()
        ast = o.to_ast()
        assert isinstance(ast, OctaveUpNode)
        assert o.to_alda() == ">"

    def test_octave_down(self):
        o = octave_down()
        ast = o.to_ast()
        assert isinstance(ast, OctaveDownNode)
        assert o.to_alda() == "<"


class TestDynamics:
    """Test dynamic markings."""

    def test_dynamics(self):
        assert pp().to_alda() == "(pp)"
        assert p().to_alda() == "(p)"
        assert mp().to_alda() == "(mp)"
        assert mf().to_alda() == "(mf)"
        assert f().to_alda() == "(f)"
        assert ff().to_alda() == "(ff)"

    def test_dynamics_to_ast(self):
        d = mf()
        ast = d.to_ast()
        assert isinstance(ast, LispListNode)


class TestOtherAttributes:
    """Test quant and panning."""

    def test_quant(self):
        q = quant(90)
        assert q.level == 90
        assert q.to_alda() == "(quant 90)"

    def test_panning(self):
        p = panning(50)
        assert p.level == 50
        assert p.to_alda() == "(panning 50)"


class TestScoreFromElements:
    """Test Score.from_elements() and related functionality."""

    def test_from_elements_basic(self):
        score = Score.from_elements(
            part("piano"), tempo(120), note("c", duration=4), note("d"), note("e")
        )
        assert score._mode == "elements"
        assert len(score._elements) == 5

    def test_from_elements_ast(self):
        score = Score.from_elements(part("piano"), note("c"), note("d"), note("e"))
        ast = score.ast
        assert ast is not None
        # Should have part declaration + event sequence
        assert len(ast.children) >= 1

    def test_from_elements_midi(self):
        score = Score.from_elements(part("piano"), note("c"), note("d"), note("e"))
        midi = score.midi
        assert midi is not None
        assert len(midi.notes) == 3

    def test_from_elements_duration(self):
        score = Score.from_elements(part("piano"), note("c", duration=4))
        assert score.duration > 0

    def test_from_elements_to_alda(self):
        score = Score.from_elements(part("piano"), note("c"), note("d"))
        alda = score.to_alda()
        assert "piano:" in alda
        assert "c" in alda
        assert "d" in alda

    def test_from_parts(self):
        score = Score.from_parts(part("piano"), part("violin"))
        assert len(score._elements) == 2

    def test_add_elements(self):
        score = Score.from_elements(part("piano"))
        score.add(note("c"), note("d"), note("e"))
        assert len(score._elements) == 4

    def test_add_returns_self(self):
        score = Score.from_elements(part("piano"))
        result = score.add(note("c"))
        assert result is score

    def test_add_to_source_raises(self):
        score = Score("piano: c d e")
        with pytest.raises(ValueError):
            score.add(note("f"))

    def test_with_part(self):
        score = Score.from_elements()
        score.with_part("piano")
        assert len(score._elements) == 1

    def test_with_tempo(self):
        score = Score.from_elements(part("piano"))
        score.with_tempo(120)
        assert len(score._elements) == 2

    def test_with_volume(self):
        score = Score.from_elements(part("piano"))
        score.with_volume(80)
        assert len(score._elements) == 2

    def test_method_chaining(self):
        score = (
            Score.from_elements()
            .with_part("piano")
            .with_tempo(120)
            .add(note("c"), note("d"), note("e"))
        )
        assert len(score._elements) == 5

    def test_repr_elements(self):
        score = Score.from_elements(part("piano"), note("c"))
        r = repr(score)
        assert "from_elements" in r
        assert "2 elements" in r

    def test_save_alda(self, tmp_path):
        score = Score.from_elements(part("piano"), note("c"))
        output = tmp_path / "test.alda"
        score.save(output)
        assert output.exists()
        content = output.read_text()
        assert "piano:" in content

    def test_save_midi(self, tmp_path):
        score = Score.from_elements(part("piano"), note("c"))
        output = tmp_path / "test.mid"
        score.save(output)
        assert output.exists()
        # MIDI files start with "MThd"
        with open(output, "rb") as f:
            assert f.read(4) == b"MThd"

    def test_part_generates_partnode_in_ast(self):
        """Regression: compose API must wrap parts in PartNode for MIDI generator."""
        score = Score.from_elements(part("violin"), note("c"), note("d"))
        ast = score.ast
        # Should have exactly one PartNode wrapping declaration + events
        part_nodes = [c for c in ast.children if isinstance(c, PartNode)]
        assert len(part_nodes) == 1
        assert part_nodes[0].declaration.names == ["violin"]
        assert len(part_nodes[0].events.events) == 2

    def test_part_instrument_honored_in_midi(self):
        """Regression: instrument from part() must affect MIDI program change."""
        score = Score.from_elements(part("violin"), note("c"))
        midi = score.midi
        # Violin should be MIDI program 40, not 0 (piano)
        assert any(pc.program == 40 for pc in midi.program_changes)

    def test_multiple_parts_generate_multiple_partnodes(self):
        """Regression: multiple parts should each get their own PartNode."""
        score = Score.from_elements(
            part("piano"),
            note("c"),
            part("violin"),
            note("d"),
        )
        ast = score.ast
        part_nodes = [c for c in ast.children if isinstance(c, PartNode)]
        assert len(part_nodes) == 2
        assert part_nodes[0].declaration.names == ["piano"]
        assert part_nodes[1].declaration.names == ["violin"]

    def test_to_alda_with_chord_no_crash(self):
        """Regression: to_alda() must not crash on chords (ChordNode has no duration)."""
        score = Score.from_elements(part("piano"), chord("c", "e", "g", duration=4))
        # This used to raise AttributeError: 'ChordNode' has no attribute 'duration'
        alda = score.to_alda()
        assert "c" in alda
        assert "e" in alda
        assert "g" in alda

    def test_to_alda_preserves_part_structure(self):
        """Regression: to_alda() must render PartNode with declaration + events."""
        score = Score.from_elements(
            part("violin"),
            note("c"),
            note("d"),
            part("piano"),
            note("e"),
            note("f"),
        )
        alda = score.to_alda()
        # Both parts should be present in output
        assert "violin:" in alda
        assert "piano:" in alda
        # Notes should be present
        assert "c" in alda
        assert "d" in alda
        assert "e" in alda
        assert "f" in alda


class TestIntegration:
    """Integration tests for compose -> MIDI pipeline."""

    def test_simple_melody(self, tmp_path):
        """Test a simple melody plays correctly."""
        score = Score.from_elements(
            part("piano"),
            tempo(120),
            note("c", duration=4),
            note("d", duration=4),
            note("e", duration=4),
            note("f", duration=4),
        )

        midi = score.midi
        assert len(midi.notes) == 4

        # Verify pitches
        pitches = [n.pitch for n in midi.notes]
        assert pitches == [60, 62, 64, 65]  # C, D, E, F

    def test_chord_progression(self, tmp_path):
        """Test chord generation."""
        score = Score.from_elements(
            part("piano"),
            chord("c", "e", "g", duration=1),
            chord("f", "a", "c", duration=1),
        )

        midi = score.midi
        assert len(midi.notes) == 6  # 3 notes per chord * 2 chords

    def test_with_attributes(self, tmp_path):
        """Test that attributes affect MIDI generation."""
        score = Score.from_elements(
            part("piano"), tempo(60), volume(80), note("c", duration=4)
        )

        midi = score.midi
        assert len(midi.notes) == 1
        # At tempo 60, quarter note = 1 second
        assert abs(midi.notes[0].duration - 1.0) < 0.2  # Allow for quantization

    def test_repeated_sequence(self, tmp_path):
        """Test sequence repetition."""
        score = Score.from_elements(
            part("piano"), seq(note("c", duration=8), note("d", duration=8)) * 4
        )

        midi = score.midi
        assert len(midi.notes) == 8  # 2 notes * 4 repetitions



"""Recursive descent parser for the Alda music programming language."""

from typing import NoReturn

from .ast_nodes import (
    ASTNode,
    AtMarkerNode,
    BarlineNode,  # Phase 2 nodes
    BracketedSequenceNode,
    ChordNode,
    CramNode,
    DurationNode,
    EventSequenceNode,
    LispListNode,
    LispNumberNode,
    LispStringNode,
    LispSymbolNode,
    MarkerNode,
    NoteLengthMsNode,
    NoteLengthNode,
    NoteLengthSecondsNode,
    NoteNode,
    OctaveDownNode,
    OctaveSetNode,
    OctaveUpNode,
    OnRepetitionsNode,
    PartDeclarationNode,
    PartNode,
    RepeatNode,
    RepetitionRange,
    RestNode,
    RootNode,
    VariableDefinitionNode,
    VariableReferenceNode,
    VoiceGroupNode,
    VoiceNode,
)
from .errors import AldaSyntaxError
from .scanner import Scanner
from .tokens import SourcePosition, Token, TokenType


class Parser:
    """Parses Alda tokens into an AST."""

    def __init__(self, tokens: list[Token]):
        self.tokens = tokens
        self._current = 0

    @classmethod
    def from_source(cls, source: str, filename: str = "<input>") -> "Parser":
        """Create a parser from source code."""
        scanner = Scanner(source, filename)
        tokens = scanner.scan()
        return cls(tokens)

    def parse(self) -> RootNode:
        """Parse the tokens and return the AST root."""
        children: list[ASTNode] = []
        self._skip_newlines()

        while not self._is_at_end():
            node = self._parse_top_level()
            if node:
                children.append(node)
            self._skip_newlines()

        return RootNode(children=children, position=SourcePosition(1, 1))

    def _parse_top_level(self) -> ASTNode | None:
        """Parse a top-level construct (part or implicit events)."""
        # Check if this is a part declaration
        if self._is_part_declaration():
            return self._parse_part()

        # Otherwise, parse as implicit events (events without explicit part)
        events = self._parse_event_sequence_content()
        if events:
            return EventSequenceNode(events=events, position=events[0].position)
        return None

    def _is_part_declaration(self) -> bool:
        """Check if current position starts a part declaration."""
        if not self._check(TokenType.NAME):
            return False

        # Look ahead for colon (part separator)
        pos = self._current
        while pos < len(self.tokens):
            token = self.tokens[pos]
            if token.type == TokenType.COLON:
                return True
            if token.type == TokenType.EQUALS:
                # This is a variable definition, not a part
                return False
            if token.type in (TokenType.SEPARATOR, TokenType.ALIAS):
                # Could be part of multi-instrument declaration
                pos += 1
                continue
            if token.type == TokenType.NAME:
                pos += 1
                continue
            break
        return False

    def _is_variable_definition(self) -> bool:
        """Check if current position starts a variable definition."""
        if not self._check(TokenType.NAME):
            return False
        # Look ahead for equals sign
        pos = self._current + 1
        if pos < len(self.tokens) and self.tokens[pos].type == TokenType.EQUALS:
            return True
        return False

    def _parse_part(self) -> PartNode:
        """Parse a part declaration and its events."""
        declaration = self._parse_part_declaration()
        self._skip_newlines()

        # Parse events until we hit another part or EOF
        events = self._parse_event_sequence_content()

        return PartNode(
            declaration=declaration,
            events=EventSequenceNode(events=events, position=declaration.position),
            position=declaration.position,
        )

    def _parse_part_declaration(self) -> PartDeclarationNode:
        """Parse a part declaration (e.g., 'piano:', 'violin/viola "strings":')."""
        position = self._peek().position
        names: list[str] = []

        # Parse first name
        name_token = self._consume(TokenType.NAME, "Expected instrument name")
        names.append(name_token.literal)

        # Parse additional names separated by /
        while self._match(TokenType.SEPARATOR):
            name_token = self._consume(
                TokenType.NAME, "Expected instrument name after '/'"
            )
            names.append(name_token.literal)

        # Optional alias
        alias = None
        if self._match(TokenType.ALIAS):
            alias = self.tokens[self._current - 1].literal

        # Consume colon
        self._consume(TokenType.COLON, "Expected ':' after part declaration")

        return PartDeclarationNode(names=names, alias=alias, position=position)

    def _parse_event_sequence_content(
        self, stop_tokens: set[TokenType] | None = None
    ) -> list[ASTNode]:
        """Parse events until we hit a stop token, part declaration, or EOF."""
        events: list[ASTNode] = []
        stop_tokens = stop_tokens or set()

        while not self._is_at_end() and not self._is_part_declaration():
            # Check for stop tokens
            if self._peek().type in stop_tokens:
                break

            self._skip_newlines()
            if self._is_at_end() or self._is_part_declaration():
                break
            if self._peek().type in stop_tokens:
                break

            event = self._parse_event()
            if event:
                events.append(event)

        return events

    def _parse_event(self) -> ASTNode | None:
        """Parse a single event."""
        # Skip newlines between events
        if self._match(TokenType.NEWLINE):
            return None

        event = self._parse_primary_event()
        if event is None:
            return None

        # Check for postfix operators (repeat, on-repetitions)
        event = self._parse_postfix(event)

        return event

    def _parse_primary_event(self) -> ASTNode | None:
        """Parse a primary event (without postfix operators)."""
        # Barline
        if self._match(TokenType.BARLINE):
            return BarlineNode(position=self.tokens[self._current - 1].position)

        # Octave controls
        if self._match(TokenType.OCTAVE_UP):
            return OctaveUpNode(position=self.tokens[self._current - 1].position)
        if self._match(TokenType.OCTAVE_DOWN):
            return OctaveDownNode(position=self.tokens[self._current - 1].position)
        if self._match(TokenType.OCTAVE_SET):
            token = self.tokens[self._current - 1]
            return OctaveSetNode(octave=token.literal, position=token.position)

        # S-expression
        if self._check(TokenType.LEFT_PAREN):
            return self._parse_sexp()

        # Markers
        if self._match(TokenType.MARKER):
            token = self.tokens[self._current - 1]
            return MarkerNode(name=token.literal, position=token.position)

        if self._match(TokenType.AT_MARKER):
            token = self.tokens[self._current - 1]
            return AtMarkerNode(name=token.literal, position=token.position)

        # Voice marker (starts a voice group)
        if self._check(TokenType.VOICE_MARKER):
            return self._parse_voice_group()

        # Cram expression
        if self._check(TokenType.CRAM_OPEN):
            return self._parse_cram()

        # Bracketed sequence
        if self._check(TokenType.EVENT_SEQ_OPEN):
            return self._parse_bracketed_sequence()

        # Variable definition or reference
        if self._is_variable_definition():
            return self._parse_variable_definition()

        if self._check(TokenType.NAME):
            return self._parse_variable_reference()

        # Rest
        if self._check(TokenType.REST_LETTER):
            return self._parse_rest()

        # Note (might become a chord)
        if self._check(TokenType.NOTE_LETTER):
            return self._parse_note_or_chord()

        # Unexpected token - skip it
        if not self._is_at_end():
            self._advance()
            return None

        return None

    def _parse_postfix(self, event: ASTNode) -> ASTNode:
        """Parse postfix operators (repeat, on-repetitions)."""
        # Check for repeat (*N)
        if self._match(TokenType.REPEAT):
            token = self.tokens[self._current - 1]
            times = token.literal
            event = RepeatNode(event=event, times=times, position=event.position)

        # Check for on-repetitions ('1-3,5)
        if self._match(TokenType.REPETITIONS):
            token = self.tokens[self._current - 1]
            ranges = self._parse_repetition_ranges(token.literal)
            event = OnRepetitionsNode(
                event=event, ranges=ranges, position=event.position
            )

        return event

    def _parse_repetition_ranges(self, ranges_str: str) -> list[RepetitionRange]:
        """Parse a repetition ranges string like '1-3,5' into RepetitionRange objects."""
        ranges = []
        for part in ranges_str.split(","):
            if "-" in part:
                first, last = part.split("-", 1)
                ranges.append(RepetitionRange(first=int(first), last=int(last)))
            else:
                ranges.append(RepetitionRange(first=int(part)))
        return ranges

    def _parse_variable_definition(self) -> VariableDefinitionNode:
        """Parse a variable definition (name = events)."""
        position = self._peek().position
        name_token = self._consume(TokenType.NAME, "Expected variable name")
        name = name_token.literal

        self._consume(TokenType.EQUALS, "Expected '=' after variable name")

        # Parse events on the same line or until newline
        events: list[ASTNode] = []
        while not self._is_at_end() and not self._check(TokenType.NEWLINE):
            if self._is_part_declaration():
                break
            event = self._parse_event()
            if event:
                events.append(event)

        return VariableDefinitionNode(
            name=name,
            events=EventSequenceNode(events=events, position=position),
            position=position,
        )

    def _parse_variable_reference(self) -> VariableReferenceNode:
        """Parse a variable reference."""
        token = self._consume(TokenType.NAME, "Expected variable name")
        return VariableReferenceNode(name=token.literal, position=token.position)

    def _parse_voice_group(self) -> VoiceGroupNode:
        """Parse a voice group (V1: events V2: events V0:)."""
        position = self._peek().position
        voices: list[VoiceNode] = []

        while self._check(TokenType.VOICE_MARKER):
            voice = self._parse_voice()
            if voice.number == 0:
                # V0: ends the voice group
                break
            voices.append(voice)

        return VoiceGroupNode(voices=voices, position=position)

    def _parse_voice(self) -> VoiceNode:
        """Parse a single voice."""
        token = self._consume(TokenType.VOICE_MARKER, "Expected voice marker")
        number = token.literal
        position = token.position

        if number == 0:
            # V0: is just an end marker, no events
            return VoiceNode(
                number=0,
                events=EventSequenceNode(events=[], position=position),
                position=position,
            )

        # Parse events until next voice marker or end of voice group context
        events = self._parse_event_sequence_content(
            stop_tokens={TokenType.VOICE_MARKER}
        )

        return VoiceNode(
            number=number,
            events=EventSequenceNode(events=events, position=position),
            position=position,
        )

    def _parse_cram(self) -> CramNode:
        """Parse a cram expression ({events}duration)."""
        position = self._peek().position
        self._consume(TokenType.CRAM_OPEN, "Expected '{'")

        # Parse events until closing brace
        events = self._parse_event_sequence_content(stop_tokens={TokenType.CRAM_CLOSE})

        self._consume(TokenType.CRAM_CLOSE, "Expected '}'")

        # Optional duration after the cram
        duration = self._try_parse_duration()

        return CramNode(
            events=EventSequenceNode(events=events, position=position),
            duration=duration,
            position=position,
        )

    def _parse_bracketed_sequence(self) -> BracketedSequenceNode:
        """Parse a bracketed event sequence ([events])."""
        position = self._peek().position
        self._consume(TokenType.EVENT_SEQ_OPEN, "Expected '['")

        # Parse events until closing bracket
        events = self._parse_event_sequence_content(
            stop_tokens={TokenType.EVENT_SEQ_CLOSE}
        )

        self._consume(TokenType.EVENT_SEQ_CLOSE, "Expected ']'")

        return BracketedSequenceNode(
            events=EventSequenceNode(events=events, position=position),
            position=position,
        )

    def _parse_note_or_chord(self) -> ASTNode:
        """Parse a note, which might be part of a chord."""
        first_note = self._parse_note()

        # Check if this is a chord (followed by / and more notes)
        if not self._check(TokenType.SEPARATOR):
            return first_note

        # Parse as chord
        chord_elements: list[
            NoteNode
            | RestNode
            | OctaveUpNode
            | OctaveDownNode
            | OctaveSetNode
            | LispListNode
        ] = [first_note]

        while self._match(TokenType.SEPARATOR):
            # Allow octave changes between chord notes
            while (
                self._check(TokenType.OCTAVE_UP)
                or self._check(TokenType.OCTAVE_DOWN)
                or self._check(TokenType.OCTAVE_SET)
                or self._check(TokenType.LEFT_PAREN)
            ):
                if self._match(TokenType.OCTAVE_UP):
                    chord_elements.append(
                        OctaveUpNode(position=self.tokens[self._current - 1].position)
                    )
                elif self._match(TokenType.OCTAVE_DOWN):
                    chord_elements.append(
                        OctaveDownNode(position=self.tokens[self._current - 1].position)
                    )
                elif self._match(TokenType.OCTAVE_SET):
                    token = self.tokens[self._current - 1]
                    chord_elements.append(
                        OctaveSetNode(octave=token.literal, position=token.position)
                    )
                elif self._check(TokenType.LEFT_PAREN):
                    chord_elements.append(self._parse_sexp())

            if self._check(TokenType.NOTE_LETTER):
                note = self._parse_note()
                chord_elements.append(note)
            elif self._check(TokenType.REST_LETTER):
                # Rests can appear in chords too
                rest = self._parse_rest()
                chord_elements.append(rest)

        return ChordNode(notes=chord_elements, position=first_note.position)

    def _parse_note(self) -> NoteNode:
        """Parse a single note with optional accidentals and duration."""
        position = self._peek().position
        letter_token = self._consume(TokenType.NOTE_LETTER, "Expected note letter")
        letter = letter_token.literal

        # Parse accidentals
        accidentals: list[str] = []
        while True:
            if self._match(TokenType.SHARP):
                accidentals.append("+")
            elif self._match(TokenType.FLAT):
                accidentals.append("-")
            elif self._match(TokenType.NATURAL):
                accidentals.append("_")
            else:
                break

        # Parse optional duration
        duration = self._try_parse_duration()

        # Check for slur/tie to next note
        slurred = self._match(TokenType.TIE)

        return NoteNode(
            letter=letter,
            accidentals=accidentals,
            duration=duration,
            slurred=slurred,
            position=position,
        )

    def _parse_rest(self) -> RestNode:
        """Parse a rest with optional duration."""
        position = self._peek().position
        self._consume(TokenType.REST_LETTER, "Expected 'r'")

        duration = self._try_parse_duration()

        return RestNode(duration=duration, position=position)

    def _try_parse_duration(self) -> DurationNode | None:
        """Try to parse a duration. Returns None if no duration present."""
        if not self._is_duration_start():
            return None

        components = []
        position = self._peek().position

        while self._is_duration_start():
            component = self._parse_duration_component()
            components.append(component)

            # Check for tie connecting to another duration component
            if self._check(TokenType.TIE) and self._is_duration_start_at(
                self._current + 1
            ):
                self._advance()  # consume tie
            else:
                break

        if not components:
            return None

        return DurationNode(components=components, position=position)

    def _is_duration_start(self) -> bool:
        """Check if current token starts a duration."""
        return (
            self._check(TokenType.NOTE_LENGTH)
            or self._check(TokenType.NOTE_LENGTH_MS)
            or self._check(TokenType.NOTE_LENGTH_SECONDS)
        )

    def _is_duration_start_at(self, pos: int) -> bool:
        """Check if token at position starts a duration."""
        if pos >= len(self.tokens):
            return False
        token = self.tokens[pos]
        return token.type in (
            TokenType.NOTE_LENGTH,
            TokenType.NOTE_LENGTH_MS,
            TokenType.NOTE_LENGTH_SECONDS,
        )

    def _parse_duration_component(self) -> ASTNode:
        """Parse a single duration component."""
        if self._match(TokenType.NOTE_LENGTH_MS):
            token = self.tokens[self._current - 1]
            return NoteLengthMsNode(ms=token.literal, position=token.position)

        if self._match(TokenType.NOTE_LENGTH_SECONDS):
            token = self.tokens[self._current - 1]
            return NoteLengthSecondsNode(seconds=token.literal, position=token.position)

        if self._match(TokenType.NOTE_LENGTH):
            token = self.tokens[self._current - 1]
            denominator = token.literal

            # Count dots
            dots = 0
            while self._match(TokenType.DOT):
                dots += 1

            return NoteLengthNode(
                denominator=denominator, dots=dots, position=token.position
            )

        self._error("Expected duration component")

    def _parse_sexp(self) -> LispListNode:
        """Parse an S-expression."""
        position = self._peek().position
        self._consume(TokenType.LEFT_PAREN, "Expected '('")

        elements = []
        while not self._check(TokenType.RIGHT_PAREN) and not self._is_at_end():
            element = self._parse_lisp_element()
            if element:
                elements.append(element)

        self._consume(TokenType.RIGHT_PAREN, "Expected ')'")

        return LispListNode(elements=elements, position=position)

    def _parse_lisp_element(self) -> ASTNode | None:
        """Parse a single Lisp element."""
        if self._match(TokenType.SYMBOL):
            token = self.tokens[self._current - 1]
            return LispSymbolNode(name=token.literal, position=token.position)

        if self._match(TokenType.NUMBER):
            token = self.tokens[self._current - 1]
            return LispNumberNode(value=token.literal, position=token.position)

        if self._match(TokenType.STRING):
            token = self.tokens[self._current - 1]
            return LispStringNode(value=token.literal, position=token.position)

        if self._check(TokenType.LEFT_PAREN):
            return self._parse_sexp()

        return None

    # Helper methods

    def _is_at_end(self) -> bool:
        return self._peek().type == TokenType.EOF

    def _peek(self) -> Token:
        return self.tokens[self._current]

    def _peek_next(self) -> Token:
        if self._current + 1 >= len(self.tokens):
            return self.tokens[-1]  # Return EOF
        return self.tokens[self._current + 1]

    def _advance(self) -> Token:
        if not self._is_at_end():
            self._current += 1
        return self.tokens[self._current - 1]

    def _check(self, token_type: TokenType) -> bool:
        if self._is_at_end():
            return False
        return self._peek().type == token_type

    def _match(self, token_type: TokenType) -> bool:
        if self._check(token_type):
            self._advance()
            return True
        return False

    def _consume(self, token_type: TokenType, message: str) -> Token:
        if self._check(token_type):
            return self._advance()
        self._error(message)

    def _skip_newlines(self) -> None:
        while self._match(TokenType.NEWLINE):
            pass

    def _error(self, message: str) -> NoReturn:
        token = self._peek()
        raise AldaSyntaxError(message, token.position)


def parse(source: str, filename: str = "<input>") -> RootNode:
    """Convenience function to parse Alda source code."""
    parser = Parser.from_source(source, filename)
    return parser.parse()



"""Standard MIDI File (SMF) writer."""

import struct
from pathlib import Path

from .types import MidiSequence


def _write_variable_length(value: int) -> bytes:
    """Encode an integer as a MIDI variable-length quantity."""
    if value < 0:
        raise ValueError("Variable-length value must be non-negative")

    if value == 0:
        return b"\x00"

    result = []
    while value:
        result.append(value & 0x7F)
        value >>= 7

    # Reverse and set continuation bits
    result.reverse()
    for i in range(len(result) - 1):
        result[i] |= 0x80

    return bytes(result)


def _seconds_to_ticks_simple(seconds: float, ticks_per_beat: int, tempo_us: int) -> int:
    """Convert seconds to MIDI ticks assuming constant tempo.

    Args:
        seconds: Time in seconds.
        ticks_per_beat: MIDI ticks per beat.
        tempo_us: Tempo in microseconds per beat.

    Returns:
        Time in MIDI ticks.
    """
    # tempo_us is microseconds per beat
    # seconds * 1_000_000 = microseconds
    # microseconds / tempo_us = beats
    # beats * ticks_per_beat = ticks
    beats = (seconds * 1_000_000) / tempo_us
    return int(beats * ticks_per_beat)


class TempoMap:
    """Maps absolute time (seconds) to MIDI ticks, accounting for tempo changes."""

    def __init__(self, sequence: "MidiSequence") -> None:
        self.ticks_per_beat = sequence.ticks_per_beat
        self.default_tempo_us = 500000  # 120 BPM

        # Build sorted list of (time_seconds, tempo_us, tick_at_this_time)
        # We precompute tick positions for each tempo change
        self._tempo_points: list[tuple[float, int, int]] = []

        if not sequence.tempo_changes:
            # No tempo changes - everything at default tempo
            self._tempo_points = [(0.0, self.default_tempo_us, 0)]
            return

        sorted_changes = sorted(sequence.tempo_changes, key=lambda t: t.time)

        # First tempo change may not be at t=0, so handle initial segment
        current_tick = 0
        current_time = 0.0
        current_tempo_us = self.default_tempo_us

        for tc in sorted_changes:
            if tc.time > current_time:
                # Compute ticks elapsed during this segment
                segment_duration = tc.time - current_time
                segment_ticks = _seconds_to_ticks_simple(
                    segment_duration, self.ticks_per_beat, current_tempo_us
                )
                current_tick += segment_ticks

            # Record this tempo change point
            new_tempo_us = _bpm_to_tempo(tc.bpm)
            self._tempo_points.append((tc.time, new_tempo_us, current_tick))
            current_time = tc.time
            current_tempo_us = new_tempo_us

        # If first tempo change wasn't at t=0, insert the initial segment
        if self._tempo_points[0][0] > 0:
            self._tempo_points.insert(0, (0.0, self.default_tempo_us, 0))

    def seconds_to_ticks(self, seconds: float) -> int:
        """Convert absolute time in seconds to MIDI ticks."""
        if seconds <= 0:
            return 0

        # Find the tempo segment containing this timestamp
        # Walk through tempo points to find the last one <= seconds
        last_point_time = 0.0
        last_point_tempo_us = self.default_tempo_us
        last_point_tick = 0

        for point_time, point_tempo_us, point_tick in self._tempo_points:
            if point_time > seconds:
                break
            last_point_time = point_time
            last_point_tempo_us = point_tempo_us
            last_point_tick = point_tick

        # Compute ticks from last tempo change point to target time
        remaining_duration = seconds - last_point_time
        remaining_ticks = _seconds_to_ticks_simple(
            remaining_duration, self.ticks_per_beat, last_point_tempo_us
        )

        return last_point_tick + remaining_ticks


def _bpm_to_tempo(bpm: float) -> int:
    """Convert BPM to microseconds per beat."""
    return int(60_000_000 / bpm)


def write_midi_file(sequence: MidiSequence, path: Path | str) -> None:
    """Write a MidiSequence to a Standard MIDI File.

    Args:
        sequence: The MIDI sequence to write.
        path: Output file path.
    """
    # Build tempo map for accurate time-to-tick conversion
    tempo_map = TempoMap(sequence)

    # Group notes by channel
    channels: dict[int, list] = {}
    for note in sequence.notes:
        if note.channel not in channels:
            channels[note.channel] = []
        channels[note.channel].append(note)

    # Add program changes to their channels
    for pc in sequence.program_changes:
        if pc.channel not in channels:
            channels[pc.channel] = []

    tracks: list[bytes] = []

    # Track 0: tempo track
    tempo_track_data = _build_tempo_track(sequence, tempo_map)
    tracks.append(tempo_track_data)

    # One track per channel
    for channel in sorted(channels.keys()):
        track_data = _build_channel_track(sequence, channel, tempo_map)
        tracks.append(track_data)

    # Build the complete file
    output = _build_header(len(tracks), sequence.ticks_per_beat)
    for track_data in tracks:
        output += _build_track_chunk(track_data)

    # Write to file
    Path(path).write_bytes(output)


def _build_header(num_tracks: int, ticks_per_beat: int) -> bytes:
    """Build the MIDI file header chunk."""
    # MThd chunk
    # Format 1: multiple tracks, synchronous
    # Format type (2 bytes) + num tracks (2 bytes) + time division (2 bytes)
    header_data = struct.pack(">HHH", 1, num_tracks, ticks_per_beat)
    return b"MThd" + struct.pack(">I", len(header_data)) + header_data


def _build_track_chunk(track_data: bytes) -> bytes:
    """Wrap track data in an MTrk chunk."""
    return b"MTrk" + struct.pack(">I", len(track_data)) + track_data


def _build_tempo_track(sequence: MidiSequence, tempo_map: TempoMap) -> bytes:
    """Build the tempo track (track 0)."""
    events: list[tuple[int, bytes]] = []

    default_tempo_us = 500000  # 120 BPM

    if sequence.tempo_changes:
        for tc in sorted(sequence.tempo_changes, key=lambda t: t.time):
            tempo_us = _bpm_to_tempo(tc.bpm)
            tick = tempo_map.seconds_to_ticks(tc.time)
            # Meta event: FF 51 03 tt tt tt (set tempo)
            tempo_bytes = struct.pack(">I", tempo_us)[1:]  # 3 bytes, big-endian
            events.append((tick, b"\xff\x51\x03" + tempo_bytes))
    else:
        # Add default tempo at time 0
        tempo_bytes = struct.pack(">I", default_tempo_us)[1:]
        events.append((0, b"\xff\x51\x03" + tempo_bytes))

    # End of track
    if events:
        last_tick = max(e[0] for e in events)
    else:
        last_tick = 0
    events.append((last_tick, b"\xff\x2f\x00"))

    return _encode_track_events(events)


def _build_channel_track(
    sequence: MidiSequence, channel: int, tempo_map: TempoMap
) -> bytes:
    """Build a track for a specific MIDI channel."""
    events: list[tuple[int, bytes]] = []

    # Add program changes
    for pc in sequence.program_changes:
        if pc.channel == channel:
            tick = tempo_map.seconds_to_ticks(pc.time)
            # Program change: Cn pp
            msg = bytes([0xC0 | (channel & 0x0F), pc.program & 0x7F])
            events.append((tick, msg))

    # Add control changes
    for cc in sequence.control_changes:
        if cc.channel == channel:
            tick = tempo_map.seconds_to_ticks(cc.time)
            # Control change: Bn cc vv
            msg = bytes([0xB0 | (channel & 0x0F), cc.control & 0x7F, cc.value & 0x7F])
            events.append((tick, msg))

    # Add note on/off events
    for note in sequence.notes:
        if note.channel == channel:
            start_tick = tempo_map.seconds_to_ticks(note.start_time)
            end_tick = tempo_map.seconds_to_ticks(note.start_time + note.duration)

            # Note on: 9n kk vv
            note_on = bytes(
                [0x90 | (channel & 0x0F), note.pitch & 0x7F, note.velocity & 0x7F]
            )
            # Note off: 8n kk vv
            note_off = bytes([0x80 | (channel & 0x0F), note.pitch & 0x7F, 0])

            events.append((start_tick, note_on))
            events.append((end_tick, note_off))

    # Sort events: by tick, then note_off before note_on at same tick
    events.sort(key=lambda e: (e[0], e[1][0] & 0xF0 != 0x80))

    # End of track
    if events:
        last_tick = max(e[0] for e in events)
    else:
        last_tick = 0
    events.append((last_tick, b"\xff\x2f\x00"))

    return _encode_track_events(events)


def _encode_track_events(events: list[tuple[int, bytes]]) -> bytes:
    """Encode a list of (absolute_tick, event_bytes) to track data with delta times."""
    result = bytearray()
    last_tick = 0

    for tick, event_data in events:
        delta = max(0, tick - last_tick)
        result.extend(_write_variable_length(delta))
        result.extend(event_data)
        last_tick = tick

    return bytes(result)



"""Base classes and protocols for compose elements."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..ast_nodes import ASTNode


class ComposeElement(ABC):
    """Base class for all compose elements.

    All compose elements can generate AST nodes directly and
    serialize to Alda source code.
    """

    @abstractmethod
    def to_ast(self) -> ASTNode:
        """Convert this element to an AST node."""
        ...

    @abstractmethod
    def to_alda(self) -> str:
        """Convert this element to Alda source code."""
        ...



from __future__ import annotations

from .key_bindings import (ConditionalKeyBindings, DynamicKeyBindings,
                           KeyBindings, KeyBindingsBase, merge_key_bindings)
from .key_processor import KeyPress, KeyPressEvent

__all__ = [
    # key_bindings.
    "ConditionalKeyBindings",
    "DynamicKeyBindings",
    "KeyBindings",
    "KeyBindingsBase",
    "merge_key_bindings",
    # key_processor
    "KeyPress",
    "KeyPressEvent",
]



"""Tests for the TinySoundFont backend."""

import os
import time
from pathlib import Path
from unittest.mock import patch

import pytest

from aldakit import Score
from aldakit.midi.types import MidiSequence, MidiNote, MidiProgramChange


# Check if TSF backend is available
try:
    from aldakit import _tsf
    from aldakit.midi.backends import (
        TsfBackend,
        HAS_TSF,
        find_soundfont,
        list_soundfonts,
    )
    from aldakit.midi.backends.tsf_backend import is_available

    TSF_AVAILABLE = HAS_TSF
except ImportError:
    TSF_AVAILABLE = False
    _tsf = None
    TsfBackend = None
    find_soundfont = None
    list_soundfonts = None

    def is_available():  # stub/fallback when TSF backend import fails
        return False


# Skip all tests if TSF not available
pytestmark = pytest.mark.skipif(
    not TSF_AVAILABLE, reason="TinySoundFont backend not available"
)


class TestTsfModule:
    """Test the native _tsf module."""

    def test_module_loads(self):
        assert _tsf is not None
        assert hasattr(_tsf, "TsfPlayer")

    def test_player_creation(self):
        player = _tsf.TsfPlayer()
        assert player is not None
        assert not player.is_loaded()
        assert not player.is_playing()

    def test_player_without_soundfont(self):
        player = _tsf.TsfPlayer()
        assert player.preset_count() == 0
        assert player.preset_name(0) == ""

    def test_player_load_invalid_path(self):
        player = _tsf.TsfPlayer()
        result = player.load_soundfont("/nonexistent/path.sf2")
        assert result is False
        assert not player.is_loaded()


class TestFindSoundFont:
    """Test SoundFont discovery functions."""

    def test_is_available(self):
        assert is_available() is True

    def test_find_soundfont_returns_path_or_none(self):
        result = find_soundfont()
        assert result is None or isinstance(result, Path)

    def test_find_soundfont_with_env_var(self, tmp_path):
        # Create a fake soundfont file
        fake_sf = tmp_path / "test.sf2"
        fake_sf.write_bytes(b"RIFF" + b"\x00" * 100)  # Minimal fake

        with patch.dict(os.environ, {"ALDAKIT_SOUNDFONT": str(fake_sf)}):
            result = find_soundfont()
            assert result == fake_sf

    def test_find_soundfont_env_var_missing_file(self):
        with patch.dict(os.environ, {"ALDAKIT_SOUNDFONT": "/nonexistent/sf.sf2"}):
            # Should fall back to searching other locations
            result = find_soundfont()
            # Result depends on system - may be None or found elsewhere
            assert result is None or isinstance(result, Path)

    def test_list_soundfonts_returns_list(self):
        result = list_soundfonts()
        assert isinstance(result, list)
        for item in result:
            assert isinstance(item, Path)


@pytest.fixture
def soundfont_path():
    """Get a valid SoundFont path, or skip if none available."""
    sf = find_soundfont()
    if sf is None:
        pytest.skip("No SoundFont available for testing")
    return sf


class TestTsfBackend:
    """Test the TsfBackend class."""

    def test_backend_init_with_soundfont(self, soundfont_path):
        backend = TsfBackend(soundfont=soundfont_path)
        assert backend.soundfont == soundfont_path
        assert backend.preset_count > 0

    def test_backend_init_auto_find(self, soundfont_path):
        # Should auto-find the SoundFont
        backend = TsfBackend()
        assert backend.soundfont is not None
        assert backend.preset_count > 0

    def test_backend_init_invalid_path(self):
        with pytest.raises(FileNotFoundError):
            TsfBackend(soundfont="/nonexistent/path.sf2")

    def test_backend_preset_names(self, soundfont_path):
        backend = TsfBackend(soundfont=soundfont_path)
        # GM SoundFonts should have piano as first preset
        name = backend.preset_name(0)
        assert isinstance(name, str)
        assert len(name) > 0

    def test_backend_set_gain(self, soundfont_path):
        backend = TsfBackend(soundfont=soundfont_path)
        backend.set_gain(0.5)
        backend.set_gain(1.5)
        # Should not raise, backend still valid
        assert backend.preset_count > 0

    def test_backend_context_manager(self, soundfont_path):
        with TsfBackend(soundfont=soundfont_path) as backend:
            assert backend is not None
            assert not backend.is_playing()

    def test_backend_repr(self, soundfont_path):
        backend = TsfBackend(soundfont=soundfont_path)
        repr_str = repr(backend)
        assert "TsfBackend" in repr_str
        assert "presets=" in repr_str


class TestTsfPlayback:
    """Test actual audio playback."""

    def test_play_simple_sequence(self, soundfont_path):
        backend = TsfBackend(soundfont=soundfont_path)

        # Create a minimal sequence
        sequence = MidiSequence(
            notes=[
                MidiNote(
                    pitch=60, velocity=100, start_time=0.0, duration=0.1, channel=0
                ),
            ],
            program_changes=[
                MidiProgramChange(program=0, time=0.0, channel=0),
            ],
        )

        backend.play(sequence)
        assert backend.is_playing()

        # Wait for completion
        backend.wait()
        assert not backend.is_playing()

    def test_play_multiple_notes(self, soundfont_path):
        backend = TsfBackend(soundfont=soundfont_path)

        # C major chord
        sequence = MidiSequence(
            notes=[
                MidiNote(
                    pitch=60, velocity=80, start_time=0.0, duration=0.2, channel=0
                ),
                MidiNote(
                    pitch=64, velocity=80, start_time=0.0, duration=0.2, channel=0
                ),
                MidiNote(
                    pitch=67, velocity=80, start_time=0.0, duration=0.2, channel=0
                ),
            ],
            program_changes=[
                MidiProgramChange(program=0, time=0.0, channel=0),
            ],
        )

        backend.play(sequence)
        backend.wait()
        assert not backend.is_playing()

    def test_play_and_stop(self, soundfont_path):
        backend = TsfBackend(soundfont=soundfont_path)

        # Longer sequence
        sequence = MidiSequence(
            notes=[
                MidiNote(
                    pitch=60, velocity=80, start_time=0.0, duration=2.0, channel=0
                ),
            ],
        )

        backend.play(sequence)
        time.sleep(0.1)
        assert backend.is_playing()

        backend.stop()
        time.sleep(0.05)
        assert not backend.is_playing()

    def test_current_time_advances(self, soundfont_path):
        backend = TsfBackend(soundfont=soundfont_path)

        sequence = MidiSequence(
            notes=[
                MidiNote(
                    pitch=60, velocity=80, start_time=0.0, duration=0.5, channel=0
                ),
            ],
        )

        backend.play(sequence)

        # Poll for time advancement with timeout (audio thread startup can be slow)
        t = 0.0
        for _ in range(20):  # Up to 1 second total
            time.sleep(0.05)
            t = backend.current_time()
            if t > 0.0:
                break

        assert t > 0.0, "current_time() never advanced from 0.0"

        backend.stop()

    def test_play_empty_sequence(self, soundfont_path):
        backend = TsfBackend(soundfont=soundfont_path)

        sequence = MidiSequence(notes=[], program_changes=[])

        backend.play(sequence)
        # Should complete quickly
        backend.wait()
        assert not backend.is_playing()


class TestScoreAudioBackend:
    """Test Score.play() with backend='audio'."""

    def test_score_play_audio_backend(self, soundfont_path):
        score = Score("piano: c8 d e")
        score.play(backend="audio", wait=True)
        assert score.duration > 0

    def test_score_play_audio_with_soundfont(self, soundfont_path):
        score = Score("piano: c4")
        score.play(backend="audio", soundfont=str(soundfont_path), wait=True)
        assert score.duration > 0

    def test_score_play_audio_chord(self, soundfont_path):
        score = Score("piano: c/e/g")
        score.play(backend="audio", wait=True)
        assert score.duration > 0

    def test_score_play_audio_multiple_parts(self, soundfont_path):
        score = Score("""
            piano: c4 d e
            violin: e4 f g
        """)
        score.play(backend="audio", wait=True)
        assert score.duration > 0

    def test_score_play_audio_with_tempo(self, soundfont_path):
        score = Score("""
            piano:
            (tempo 200)
            c8 d e f g a b > c
        """)
        score.play(backend="audio", wait=True)
        assert score.duration > 0


class TestTsfBackendSave:
    """Test TsfBackend.save() method."""

    def test_save_writes_midi_file(self, soundfont_path, tmp_path):
        backend = TsfBackend(soundfont=soundfont_path)

        sequence = MidiSequence(
            notes=[
                MidiNote(
                    pitch=60, velocity=80, start_time=0.0, duration=0.5, channel=0
                ),
            ],
        )

        output_path = tmp_path / "test_output.mid"
        backend.save(sequence, output_path)

        assert output_path.exists()
        assert output_path.stat().st_size > 0

        # Verify it's a valid MIDI file (starts with MThd)
        with open(output_path, "rb") as f:
            header = f.read(4)
            assert header == b"MThd"


class TestTsfBackendNoSoundFont:
    """Test behavior when no SoundFont is available."""

    def test_init_raises_without_soundfont(self):
        # Temporarily hide all SoundFonts
        with patch(
            "aldakit.midi.backends.tsf_backend.find_soundfont", return_value=None
        ):
            with pytest.raises(FileNotFoundError) as exc_info:
                TsfBackend()
            assert "No SoundFont" in str(exc_info.value)

    def test_score_play_audio_raises_without_soundfont(self):
        with patch(
            "aldakit.midi.backends.tsf_backend.find_soundfont", return_value=None
        ):
            score = Score("piano: c")
            with pytest.raises(FileNotFoundError):
                score.play(backend="audio")



"""MIDI backend implementations."""

from .base import MidiBackend
from .libremidi_backend import LibremidiBackend

# TsfBackend is optional (requires _tsf native module and SoundFont)
try:
    from .tsf_backend import TsfBackend, find_soundfont, list_soundfonts

    HAS_TSF = True
except ImportError:
    HAS_TSF = False
    TsfBackend = None  # type: ignore
    find_soundfont = None  # type: ignore
    list_soundfonts = None  # type: ignore

__all__ = [
    "MidiBackend",
    "LibremidiBackend",
    "TsfBackend",
    "HAS_TSF",
    "find_soundfont",
    "list_soundfonts",
]



"""Demo of the aldakit parser."""

from aldakit.parser import parse


def main():
    # Phase 1 example
    source1 = """
    piano:
      (tempo 120)
      o4 c8 d e f | g4 a b > c
    """

    print("=== Phase 1: Basic melody ===")
    ast = parse(source1)
    print(ast)
    print()

    # Phase 2 example with variables, markers, voices, cram, and repeats
    source2 = """
    # Define a motif
    theme = c4 d e f

    piano:
      %intro
      theme*2

      V1: c4 e g > c
      V2: e4 g > c e
      V0:

      {c d e}2

      [c d e f]*4

      @intro
    """

    print("=== Phase 2: Variables, voices, cram, repeats ===")
    ast2 = parse(source2)
    print(ast2)


if __name__ == "__main__":
    main()



"""Chord voicing utilities for building common chord types.

This module provides convenient constructors for building chords by name,
such as major, minor, diminished, augmented, and various seventh chords.
"""

from __future__ import annotations

from .core import Chord, Note, note
from .scales import OFFSET_TO_PITCH, PITCH_TO_OFFSET

# =============================================================================
# Chord Interval Definitions
# =============================================================================

# Intervals from root (in semitones) for common chord types
CHORD_INTERVALS: dict[str, tuple[int, ...]] = {
    # Triads
    "major": (0, 4, 7),
    "minor": (0, 3, 7),
    "diminished": (0, 3, 6),
    "augmented": (0, 4, 8),
    "sus2": (0, 2, 7),
    "sus4": (0, 5, 7),
    # Seventh chords
    "major7": (0, 4, 7, 11),
    "minor7": (0, 3, 7, 10),
    "dominant7": (0, 4, 7, 10),
    "diminished7": (0, 3, 6, 9),
    "half-diminished7": (0, 3, 6, 10),
    "minor-major7": (0, 3, 7, 11),
    "augmented7": (0, 4, 8, 10),
    "augmented-major7": (0, 4, 8, 11),
    # Sixth chords
    "major6": (0, 4, 7, 9),
    "minor6": (0, 3, 7, 9),
    # Extended chords
    "dominant9": (0, 4, 7, 10, 14),
    "major9": (0, 4, 7, 11, 14),
    "minor9": (0, 3, 7, 10, 14),
    "dominant11": (0, 4, 7, 10, 14, 17),
    "dominant13": (0, 4, 7, 10, 14, 17, 21),
    # Added tone chords
    "add9": (0, 4, 7, 14),
    "add11": (0, 4, 7, 17),
    "madd9": (0, 3, 7, 14),
    # Power chord
    "power": (0, 7),
    "5": (0, 7),
}


# =============================================================================
# Core Chord Builder
# =============================================================================


def build_chord(
    root: str,
    chord_type: str,
    *,
    octave: int = 4,
    duration: int | None = None,
    inversion: int = 0,
) -> Chord:
    """Build a chord from a root note and chord type.

    Args:
        root: Root note of the chord (e.g., "c", "f", "b").
        chord_type: Type of chord (e.g., "major", "minor7", "diminished").
        octave: Octave for the root note.
        duration: Duration for the chord.
        inversion: Chord inversion (0 = root position, 1 = first inversion, etc.).

    Returns:
        A Chord instance.

    Examples:
        >>> build_chord("c", "major")  # C major: C E G
        >>> build_chord("a", "minor7", duration=2)  # A minor 7
        >>> build_chord("g", "dominant7", inversion=1)  # G7 first inversion
    """
    if chord_type not in CHORD_INTERVALS:
        available = ", ".join(sorted(CHORD_INTERVALS.keys()))
        raise ValueError(f"Unknown chord type: {chord_type}. Available: {available}")

    # Parse root note
    if len(root) > 1 and root[1] in "+-":
        base_pitch = root[0].lower()
        root_acc = root[1]
        root_offset = PITCH_TO_OFFSET[base_pitch]
        if root_acc == "+":
            root_offset += 1
        elif root_acc == "-":
            root_offset -= 1
        root_offset = root_offset % 12
    else:
        base_pitch = root.lower()
        root_offset = PITCH_TO_OFFSET[base_pitch]
        root_acc = None

    intervals = list(CHORD_INTERVALS[chord_type])

    # Apply inversion by rotating intervals and adjusting octaves
    if inversion > 0:
        inversion = inversion % len(intervals)
        # Move lower notes up an octave
        for i in range(inversion):
            intervals[i] += 12
        # Sort to maintain ascending order
        intervals.sort()

    # Build notes
    notes = []
    for interval in intervals:
        pitch_offset = (root_offset + interval) % 12
        octave_offset = (root_offset + interval) // 12
        pitch_name, accidental = OFFSET_TO_PITCH[pitch_offset]
        notes.append(
            note(
                pitch_name,
                octave=octave + octave_offset,
                accidental=accidental,
                duration=None,  # Chord duration is set on the Chord itself
            )
        )

    return Chord(notes=tuple(notes), duration=duration)


# =============================================================================
# Triad Constructors
# =============================================================================


def major(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
    inversion: int = 0,
) -> Chord:
    """Build a major triad.

    Args:
        root: Root note.
        octave: Octave for root.
        duration: Chord duration.
        inversion: Chord inversion.

    Returns:
        Major triad chord.

    Examples:
        >>> major("c")  # C E G
        >>> major("g", duration=2)
    """
    return build_chord(
        root, "major", octave=octave, duration=duration, inversion=inversion
    )


def minor(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
    inversion: int = 0,
) -> Chord:
    """Build a minor triad.

    Examples:
        >>> minor("a")  # A C E
        >>> minor("d", inversion=1)  # First inversion
    """
    return build_chord(
        root, "minor", octave=octave, duration=duration, inversion=inversion
    )


def dim(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
    inversion: int = 0,
) -> Chord:
    """Build a diminished triad.

    Examples:
        >>> dim("b")  # B D F
    """
    return build_chord(
        root, "diminished", octave=octave, duration=duration, inversion=inversion
    )


def aug(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
    inversion: int = 0,
) -> Chord:
    """Build an augmented triad.

    Examples:
        >>> aug("c")  # C E G#
    """
    return build_chord(
        root, "augmented", octave=octave, duration=duration, inversion=inversion
    )


def sus2(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
) -> Chord:
    """Build a sus2 chord.

    Examples:
        >>> sus2("c")  # C D G
    """
    return build_chord(root, "sus2", octave=octave, duration=duration)


def sus4(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
) -> Chord:
    """Build a sus4 chord.

    Examples:
        >>> sus4("c")  # C F G
    """
    return build_chord(root, "sus4", octave=octave, duration=duration)


# =============================================================================
# Seventh Chord Constructors
# =============================================================================


def maj7(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
    inversion: int = 0,
) -> Chord:
    """Build a major seventh chord.

    Examples:
        >>> maj7("c")  # C E G B
    """
    return build_chord(
        root, "major7", octave=octave, duration=duration, inversion=inversion
    )


def min7(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
    inversion: int = 0,
) -> Chord:
    """Build a minor seventh chord.

    Examples:
        >>> min7("a")  # A C E G
    """
    return build_chord(
        root, "minor7", octave=octave, duration=duration, inversion=inversion
    )


def dom7(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
    inversion: int = 0,
) -> Chord:
    """Build a dominant seventh chord.

    Examples:
        >>> dom7("g")  # G B D F
    """
    return build_chord(
        root, "dominant7", octave=octave, duration=duration, inversion=inversion
    )


def dim7(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
    inversion: int = 0,
) -> Chord:
    """Build a fully diminished seventh chord.

    Examples:
        >>> dim7("b")  # B D F Ab
    """
    return build_chord(
        root, "diminished7", octave=octave, duration=duration, inversion=inversion
    )


def half_dim7(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
    inversion: int = 0,
) -> Chord:
    """Build a half-diminished seventh chord (minor 7 flat 5).

    Examples:
        >>> half_dim7("b")  # B D F A
    """
    return build_chord(
        root, "half-diminished7", octave=octave, duration=duration, inversion=inversion
    )


def min_maj7(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
    inversion: int = 0,
) -> Chord:
    """Build a minor-major seventh chord.

    Examples:
        >>> min_maj7("c")  # C Eb G B
    """
    return build_chord(
        root, "minor-major7", octave=octave, duration=duration, inversion=inversion
    )


def aug7(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
    inversion: int = 0,
) -> Chord:
    """Build an augmented seventh chord.

    Examples:
        >>> aug7("c")  # C E G# Bb
    """
    return build_chord(
        root, "augmented7", octave=octave, duration=duration, inversion=inversion
    )


# =============================================================================
# Sixth Chord Constructors
# =============================================================================


def maj6(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
) -> Chord:
    """Build a major sixth chord.

    Examples:
        >>> maj6("c")  # C E G A
    """
    return build_chord(root, "major6", octave=octave, duration=duration)


def min6(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
) -> Chord:
    """Build a minor sixth chord.

    Examples:
        >>> min6("a")  # A C E F#
    """
    return build_chord(root, "minor6", octave=octave, duration=duration)


# =============================================================================
# Extended Chord Constructors
# =============================================================================


def dom9(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
) -> Chord:
    """Build a dominant ninth chord.

    Examples:
        >>> dom9("g")  # G B D F A
    """
    return build_chord(root, "dominant9", octave=octave, duration=duration)


def maj9(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
) -> Chord:
    """Build a major ninth chord.

    Examples:
        >>> maj9("c")  # C E G B D
    """
    return build_chord(root, "major9", octave=octave, duration=duration)


def min9(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
) -> Chord:
    """Build a minor ninth chord.

    Examples:
        >>> min9("a")  # A C E G B
    """
    return build_chord(root, "minor9", octave=octave, duration=duration)


def add9(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
) -> Chord:
    """Build an add9 chord (major triad plus 9th, no 7th).

    Examples:
        >>> add9("c")  # C E G D
    """
    return build_chord(root, "add9", octave=octave, duration=duration)


def power(
    root: str,
    *,
    octave: int = 4,
    duration: int | None = None,
) -> Chord:
    """Build a power chord (root and fifth only).

    Examples:
        >>> power("e")  # E B
    """
    return build_chord(root, "power", octave=octave, duration=duration)


# =============================================================================
# Chord Utilities
# =============================================================================


def arpeggiate(
    chord: Chord,
    pattern: list[int] | None = None,
    *,
    duration: int | None = None,
) -> list[Note]:
    """Arpeggiate a chord into individual notes.

    Args:
        chord: The chord to arpeggiate.
        pattern: List of indices into the chord notes (0-based).
            If None, plays notes in order.
        duration: Duration for each note.

    Returns:
        List of notes forming the arpeggio.

    Examples:
        >>> c_maj = major("c")
        >>> arpeggiate(c_maj, [0, 1, 2, 1])  # C E G E
        >>> arpeggiate(c_maj, duration=16)  # C E G as 16th notes
    """
    notes = list(chord.notes)

    if pattern is None:
        pattern = list(range(len(notes)))

    result = []
    for idx in pattern:
        n = notes[idx % len(notes)]
        result.append(
            note(
                n.pitch,
                duration=duration if duration else n.duration,
                octave=n.octave,
                accidental=n.accidental,
            )
        )

    return result


def invert(chord: Chord, inversion: int) -> Chord:
    """Return an inverted version of a chord.

    Args:
        chord: The chord to invert.
        inversion: Inversion number (1 = first, 2 = second, etc.).

    Returns:
        Inverted chord.

    Examples:
        >>> c_maj = major("c")
        >>> invert(c_maj, 1)  # First inversion: E G C
    """
    notes = list(chord.notes)
    n = len(notes)

    if inversion <= 0:
        return chord

    inversion = inversion % n

    # Move lower notes up an octave
    new_notes = []
    for i, nt in enumerate(notes):
        if i < inversion:
            new_octave = (nt.octave or 4) + 1
            new_notes.append(
                note(
                    nt.pitch,
                    duration=nt.duration,
                    octave=new_octave,
                    accidental=nt.accidental,
                )
            )
        else:
            new_notes.append(nt)

    # Sort by pitch to maintain ascending order
    new_notes.sort(key=lambda x: x.midi_pitch)

    return Chord(notes=tuple(new_notes), duration=chord.duration)


def voicing(
    chord: Chord,
    octaves: list[int],
) -> Chord:
    """Apply specific octave voicing to a chord.

    Args:
        chord: The chord to voice.
        octaves: List of octaves for each note (must match chord size).

    Returns:
        Chord with new voicing.

    Examples:
        >>> c_maj = major("c")
        >>> voicing(c_maj, [3, 4, 5])  # C3 E4 G5 (spread voicing)
    """
    notes = list(chord.notes)

    if len(octaves) != len(notes):
        raise ValueError(f"octaves list must have {len(notes)} elements")

    new_notes = []
    for nt, oct in zip(notes, octaves):
        new_notes.append(
            note(
                nt.pitch,
                duration=nt.duration,
                octave=oct,
                accidental=nt.accidental,
            )
        )

    return Chord(notes=tuple(new_notes), duration=chord.duration)


def list_chord_types() -> list[str]:
    """Get a list of all available chord types.

    Returns:
        Sorted list of chord type names.

    Examples:
        >>> "major7" in list_chord_types()
        True
    """
    return sorted(CHORD_INTERVALS.keys())



"""Error types for the Alda parser."""

from .tokens import SourcePosition


class AldaParseError(Exception):
    """Error during parsing or scanning of Alda source code."""

    def __init__(
        self,
        message: str,
        position: SourcePosition | None = None,
        source_line: str | None = None,
    ):
        self.message = message
        self.position = position
        self.source_line = source_line
        super().__init__(self._format_message())

    def _format_message(self) -> str:
        parts = []

        if self.position:
            parts.append(f"{self.position}: ")

        parts.append(self.message)

        if self.source_line is not None and self.position:
            parts.append(f"\n  {self.source_line}")
            # Add caret pointing to error column
            parts.append(f"\n  {' ' * (self.position.column - 1)}^")

        return "".join(parts)


class AldaScanError(AldaParseError):
    """Error during lexical analysis."""

    pass


class AldaSyntaxError(AldaParseError):
    """Error during parsing (syntax error)."""

    pass



"""
Input validation for a `Buffer`.
(Validators will be called before accepting input.)
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Callable

from prompt_toolkit.eventloop import run_in_executor_with_context

from .document import Document
from .filters import FilterOrBool, to_filter

__all__ = [
    "ConditionalValidator",
    "ValidationError",
    "Validator",
    "ThreadedValidator",
    "DummyValidator",
    "DynamicValidator",
]


class ValidationError(Exception):
    """
    Error raised by :meth:`.Validator.validate`.

    :param cursor_position: The cursor position where the error occurred.
    :param message: Text.
    """

    def __init__(self, cursor_position: int = 0, message: str = "") -> None:
        super().__init__(message)
        self.cursor_position = cursor_position
        self.message = message

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(cursor_position={self.cursor_position!r}, message={self.message!r})"


class Validator(metaclass=ABCMeta):
    """
    Abstract base class for an input validator.

    A validator is typically created in one of the following two ways:

    - Either by overriding this class and implementing the `validate` method.
    - Or by passing a callable to `Validator.from_callable`.

    If the validation takes some time and needs to happen in a background
    thread, this can be wrapped in a :class:`.ThreadedValidator`.
    """

    @abstractmethod
    def validate(self, document: Document) -> None:
        """
        Validate the input.
        If invalid, this should raise a :class:`.ValidationError`.

        :param document: :class:`~prompt_toolkit.document.Document` instance.
        """
        pass

    async def validate_async(self, document: Document) -> None:
        """
        Return a `Future` which is set when the validation is ready.
        This function can be overloaded in order to provide an asynchronous
        implementation.
        """
        try:
            self.validate(document)
        except ValidationError:
            raise

    @classmethod
    def from_callable(
        cls,
        validate_func: Callable[[str], bool],
        error_message: str = "Invalid input",
        move_cursor_to_end: bool = False,
    ) -> Validator:
        """
        Create a validator from a simple validate callable. E.g.:

        .. code:: python

            def is_valid(text):
                return text in ['hello', 'world']
            Validator.from_callable(is_valid, error_message='Invalid input')

        :param validate_func: Callable that takes the input string, and returns
            `True` if the input is valid input.
        :param error_message: Message to be displayed if the input is invalid.
        :param move_cursor_to_end: Move the cursor to the end of the input, if
            the input is invalid.
        """
        return _ValidatorFromCallable(validate_func, error_message, move_cursor_to_end)


class _ValidatorFromCallable(Validator):
    """
    Validate input from a simple callable.
    """

    def __init__(
        self, func: Callable[[str], bool], error_message: str, move_cursor_to_end: bool
    ) -> None:
        self.func = func
        self.error_message = error_message
        self.move_cursor_to_end = move_cursor_to_end

    def __repr__(self) -> str:
        return f"Validator.from_callable({self.func!r})"

    def validate(self, document: Document) -> None:
        if not self.func(document.text):
            if self.move_cursor_to_end:
                index = len(document.text)
            else:
                index = 0

            raise ValidationError(cursor_position=index, message=self.error_message)


class ThreadedValidator(Validator):
    """
    Wrapper that runs input validation in a thread.
    (Use this to prevent the user interface from becoming unresponsive if the
    input validation takes too much time.)
    """

    def __init__(self, validator: Validator) -> None:
        self.validator = validator

    def validate(self, document: Document) -> None:
        self.validator.validate(document)

    async def validate_async(self, document: Document) -> None:
        """
        Run the `validate` function in a thread.
        """

        def run_validation_thread() -> None:
            return self.validate(document)

        await run_in_executor_with_context(run_validation_thread)


class DummyValidator(Validator):
    """
    Validator class that accepts any input.
    """

    def validate(self, document: Document) -> None:
        pass  # Don't raise any exception.


class ConditionalValidator(Validator):
    """
    Validator that can be switched on/off according to
    a filter. (This wraps around another validator.)
    """

    def __init__(self, validator: Validator, filter: FilterOrBool) -> None:
        self.validator = validator
        self.filter = to_filter(filter)

    def validate(self, document: Document) -> None:
        # Call the validator only if the filter is active.
        if self.filter():
            self.validator.validate(document)


class DynamicValidator(Validator):
    """
    Validator class that can dynamically returns any Validator.

    :param get_validator: Callable that returns a :class:`.Validator` instance.
    """

    def __init__(self, get_validator: Callable[[], Validator | None]) -> None:
        self.get_validator = get_validator

    def validate(self, document: Document) -> None:
        validator = self.get_validator() or DummyValidator()
        validator.validate(document)

    async def validate_async(self, document: Document) -> None:
        validator = self.get_validator() or DummyValidator()
        await validator.validate_async(document)



"""Tests for the high-level aldakit API."""

from unittest.mock import patch, MagicMock

import pytest

import aldakit
from aldakit import Score
from aldakit.ast_nodes import RootNode
from aldakit.midi.types import MidiSequence
from aldakit.errors import AldaParseError


class TestScore:
    """Test Score class."""

    def test_create_from_string(self):
        score = Score("piano: c d e")
        assert score.source == "piano: c d e"

    def test_create_from_file(self, tmp_path):
        alda_file = tmp_path / "test.alda"
        alda_file.write_text("piano: c d e f g")

        score = Score.from_file(alda_file)
        assert score.source == "piano: c d e f g"

    def test_from_file_not_found(self):
        with pytest.raises(FileNotFoundError):
            Score.from_file("/nonexistent/file.alda")

    def test_ast_property(self):
        score = Score("piano: c d e")
        ast = score.ast
        assert isinstance(ast, RootNode)
        assert len(ast.children) == 1

    def test_ast_cached(self):
        score = Score("piano: c d e")
        ast1 = score.ast
        ast2 = score.ast
        assert ast1 is ast2  # Same object (cached)

    def test_midi_property(self):
        score = Score("piano: c d e")
        midi = score.midi
        assert isinstance(midi, MidiSequence)
        assert len(midi.notes) == 3

    def test_midi_cached(self):
        score = Score("piano: c d e")
        midi1 = score.midi
        midi2 = score.midi
        assert midi1 is midi2  # Same object (cached)

    def test_duration_property(self):
        score = Score("piano: c4 d4 e4")  # Three quarter notes
        duration = score.duration
        assert duration > 0

    def test_parse_error(self):
        score = Score("piano: c d e (invalid")
        with pytest.raises(AldaParseError):
            _ = score.ast

    def test_repr(self):
        score = Score("piano: c d e")
        repr_str = repr(score)
        assert "Score(" in repr_str
        assert "piano: c d e" in repr_str

    def test_repr_truncates_long_source(self):
        long_source = "piano: " + "c d e f g a b " * 10
        score = Score(long_source)
        repr_str = repr(score)
        assert "..." in repr_str
        assert len(repr_str) < 100

    @patch("aldakit.score.LibremidiBackend")
    def test_play(self, mock_backend_class):
        mock_backend = MagicMock()
        mock_backend.is_playing.return_value = False
        mock_backend.__enter__ = MagicMock(return_value=mock_backend)
        mock_backend.__exit__ = MagicMock(return_value=None)
        mock_backend_class.return_value = mock_backend

        score = Score("piano: c d e")
        score.play()

        assert mock_backend_class.call_count == 1
        assert mock_backend_class.call_args.kwargs == {"port_name": None}
        assert mock_backend.play.call_count == 1

    @patch("aldakit.score.LibremidiBackend")
    def test_play_with_port(self, mock_backend_class):
        mock_backend = MagicMock()
        mock_backend.is_playing.return_value = False
        mock_backend.__enter__ = MagicMock(return_value=mock_backend)
        mock_backend.__exit__ = MagicMock(return_value=None)
        mock_backend_class.return_value = mock_backend

        score = Score("piano: c d e")
        score.play(port="TestPort")

        assert mock_backend_class.call_count == 1
        assert mock_backend_class.call_args.kwargs == {"port_name": "TestPort"}
        assert mock_backend.play.call_count == 1

    @patch("aldakit.score.write_midi_file")
    @patch("aldakit.score.LibremidiBackend")
    def test_save(self, mock_backend_class, mock_write):
        score = Score("piano: c d e")
        score.save("output.mid")

        mock_backend_class.assert_not_called()
        mock_write.assert_called_once()
        from pathlib import Path

        assert mock_write.call_args[0][1] == Path("output.mid")


class TestModuleFunctions:
    """Test module-level convenience functions."""

    @patch("aldakit.api.Score")
    def test_play(self, mock_score_class):
        mock_score = MagicMock()
        mock_score_class.return_value = mock_score

        aldakit.play("piano: c d e")

        assert mock_score_class.call_count == 1
        assert mock_score_class.call_args.args == ("piano: c d e",)
        assert mock_score.play.call_count == 1
        assert mock_score.play.call_args.kwargs == {"port": None, "wait": True}

    @patch("aldakit.api.Score")
    def test_play_with_options(self, mock_score_class):
        mock_score = MagicMock()
        mock_score_class.return_value = mock_score

        aldakit.play("piano: c d e", port="TestPort", wait=False)

        assert mock_score.play.call_count == 1
        assert mock_score.play.call_args.kwargs == {"port": "TestPort", "wait": False}

    @patch("aldakit.api.Score")
    def test_play_file(self, mock_score_class):
        mock_score = MagicMock()
        mock_score_class.from_file.return_value = mock_score

        aldakit.play_file("song.alda")

        assert mock_score_class.from_file.call_count == 1
        assert mock_score.play.call_count == 1
        assert mock_score.play.call_args.kwargs == {"port": None, "wait": True}

    @patch("aldakit.api.Score")
    def test_save(self, mock_score_class):
        mock_score = MagicMock()
        mock_score_class.return_value = mock_score

        aldakit.save("piano: c d e", "output.mid")

        mock_score_class.assert_called_once_with("piano: c d e")
        mock_score.save.assert_called_once_with("output.mid")

    @patch("aldakit.api.Score")
    def test_save_file(self, mock_score_class):
        mock_score = MagicMock()
        mock_score_class.from_file.return_value = mock_score

        aldakit.save_file("song.alda", "output.mid")

        assert mock_score_class.from_file.call_count == 1
        assert mock_score.save.call_count == 1
        assert mock_score.save.call_args.args == ("output.mid",)

    @patch("aldakit.api.LibremidiBackend")
    def test_list_ports(self, mock_backend_class):
        mock_backend = MagicMock()
        mock_backend.list_output_ports.return_value = ["Port1", "Port2"]
        mock_backend_class.return_value = mock_backend

        ports = aldakit.list_ports()

        assert ports == ["Port1", "Port2"]


class TestIntegration:
    """Integration tests using real parsing and MIDI generation."""

    def test_score_end_to_end(self, tmp_path):
        """Test full workflow: create score, access properties, save to file."""
        score = Score("""
        piano:
          (tempo 120)
          c4 d e f | g a b > c
        """)

        # Access all properties
        assert score.source.strip().startswith("piano:")
        assert isinstance(score.ast, RootNode)
        assert isinstance(score.midi, MidiSequence)
        assert score.duration > 0

        # Save to file (uses real backend for file writing)
        output_file = tmp_path / "test_output.mid"
        score.save(output_file)

        assert output_file.exists()
        assert output_file.stat().st_size > 0

    def test_save_function_creates_file(self, tmp_path):
        """Test save() function creates a valid MIDI file."""
        output_file = tmp_path / "test.mid"
        aldakit.save("piano: c d e", output_file)

        assert output_file.exists()
        # MIDI files start with "MThd"
        with open(output_file, "rb") as f:
            assert f.read(4) == b"MThd"



"""High-level Score class for working with Alda music."""

from __future__ import annotations

import time
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any

from .ast_nodes import EventSequenceNode, PartNode, RootNode
from .midi.backends import LibremidiBackend
from .midi.generator import generate_midi
from .midi.smf import write_midi_file
from .parser import parse

if TYPE_CHECKING:
    from .compose.base import ComposeElement
    from .midi.types import MidiSequence


# Mode constants for internal state
_MODE_SOURCE = "source"
_MODE_ELEMENTS = "elements"
_MODE_MIDI = "midi"


def _ast_to_alda(ast: RootNode) -> str:
    """Convert an AST back to Alda source code."""
    from .ast_nodes import (
        ChordNode,
        DurationNode,
        LispListNode,
        LispNumberNode,
        LispSymbolNode,
        NoteLengthNode,
        NoteNode,
        OctaveDownNode,
        OctaveSetNode,
        OctaveUpNode,
        PartDeclarationNode,
        PartNode as PartNodeType,
        RestNode,
    )

    def duration_to_str(d: DurationNode | None) -> str:
        if d is None:
            return ""
        # DurationNode has components, typically NoteLengthNode
        if not d.components:
            return ""
        comp = d.components[0]
        if isinstance(comp, NoteLengthNode):
            result = str(int(comp.denominator))
            result += "." * comp.dots
            return result
        return ""

    def node_to_str(node) -> str:
        if isinstance(node, PartNodeType):
            # PartNode wraps declaration + events
            decl_str = node_to_str(node.declaration)
            events_str = node_to_str(node.events)
            return f"{decl_str} {events_str}"

        elif isinstance(node, PartDeclarationNode):
            instruments = "/".join(node.names)
            return f"\n{instruments}:\n"

        elif isinstance(node, NoteNode):
            result = node.letter
            result += "".join(node.accidentals)
            result += duration_to_str(node.duration)
            return result

        elif isinstance(node, RestNode):
            result = "r"
            result += duration_to_str(node.duration)
            return result

        elif isinstance(node, ChordNode):
            # Notes carry their own durations; ChordNode has no duration attr
            notes = "/".join(node_to_str(n) for n in node.notes)
            return notes

        elif isinstance(node, LispListNode):
            parts = []
            for elem in node.elements:
                if isinstance(elem, LispSymbolNode):
                    parts.append(elem.name)
                elif isinstance(elem, LispNumberNode):
                    parts.append(str(elem.value))
                else:
                    parts.append(node_to_str(elem))
            return "(" + " ".join(parts) + ")"

        elif isinstance(node, OctaveSetNode):
            return f"o{node.octave}"

        elif isinstance(node, OctaveUpNode):
            return ">"

        elif isinstance(node, OctaveDownNode):
            return "<"

        elif hasattr(node, "events"):
            # EventSequenceNode
            return " ".join(node_to_str(e) for e in node.events)

        elif hasattr(node, "children"):
            # RootNode or similar container
            return " ".join(node_to_str(c) for c in node.children)

        else:
            return ""

    result = node_to_str(ast)
    # Clean up extra whitespace
    lines = [line.strip() for line in result.split("\n")]
    return "\n".join(line for line in lines if line)


class Score:
    """A unified score class for parsing, building, and playing music.

    The Score class provides a high-level interface for working with Alda music.
    It supports multiple construction methods:

    - From Alda source code: `Score("piano: c d e")` or `Score.from_source(...)`
    - From Alda file: `Score.from_file("song.alda")`
    - From compose elements: `Score.from_elements(part("piano"), note("c"), ...)`

    Examples:
        >>> # From source code
        >>> score = Score("piano: c d e f g")
        >>> score.play()

        >>> # From file
        >>> score = Score.from_file("song.alda")
        >>> score.save("output.mid")

        >>> # From compose elements
        >>> from aldakit.compose import part, note, tempo
        >>> score = Score.from_elements(
        ...     part("piano"),
        ...     tempo(120),
        ...     note("c", duration=4),
        ...     note("d"),
        ...     note("e"),
        ... )
        >>> score.play()

        >>> # Builder pattern
        >>> score = Score.from_elements(part("piano"))
        >>> score.add(note("c"), note("d"), note("e"))
        >>> score.play()
    """

    def __init__(self, source: str, filename: str = "<input>") -> None:
        """Create a Score from Alda source code.

        Args:
            source: Alda source code string.
            filename: Optional filename for error messages.
        """
        self._mode = _MODE_SOURCE
        self._source = source
        self._filename = filename
        self._elements: list[ComposeElement] = []
        self._imported_ast: RootNode | None = None

    @classmethod
    def from_source(cls, source: str, filename: str = "<input>") -> Score:
        """Create a Score from Alda source code.

        This is equivalent to calling Score(source, filename) directly.

        Args:
            source: Alda source code string.
            filename: Optional filename for error messages.

        Returns:
            A new Score instance.
        """
        return cls(source, filename)

    @classmethod
    def from_file(cls, path: str | Path) -> Score:
        """Create a Score from an Alda or MIDI file.

        Args:
            path: Path to the Alda (.alda) or MIDI (.mid, .midi) file.

        Returns:
            A new Score instance.

        Raises:
            FileNotFoundError: If the file does not exist.
            ValueError: If the file type is not supported.
        """
        path = Path(path)

        if path.suffix.lower() in (".mid", ".midi"):
            return cls.from_midi_file(path)
        elif path.suffix.lower() == ".alda":
            source = path.read_text(encoding="utf-8")
            return cls(source, filename=str(path))
        else:
            # Try to read as Alda source
            source = path.read_text(encoding="utf-8")
            return cls(source, filename=str(path))

    @classmethod
    def from_midi_file(
        cls,
        path: str | Path,
        *,
        quantize_grid: float = 0.25,
    ) -> Score:
        """Create a Score by importing a MIDI file.

        This converts the MIDI file to an AST representation, allowing
        it to be played, exported to Alda, or further manipulated.

        Args:
            path: Path to the MIDI file.
            quantize_grid: Grid size in beats for quantization (0.25 = 16th notes).
                Set to 0 to disable quantization.

        Returns:
            A new Score instance.

        Raises:
            FileNotFoundError: If the file does not exist.
            MidiParseError: If the MIDI file cannot be parsed.

        Examples:
            >>> score = Score.from_midi_file("recording.mid")
            >>> print(score.to_alda())
            >>> score.play()
        """
        from .midi.midi_to_ast import midi_to_ast
        from .midi.smf_reader import read_midi_file

        path = Path(path)
        midi_sequence = read_midi_file(path)
        ast = midi_to_ast(midi_sequence, quantize_grid=quantize_grid)

        score = cls.__new__(cls)
        score._mode = _MODE_MIDI
        score._source = ""
        score._filename = str(path)
        score._elements = []
        score._imported_ast = ast
        return score

    @classmethod
    def from_elements(cls, *elements: ComposeElement) -> Score:
        """Create a Score from compose domain objects.

        Args:
            *elements: Compose elements (notes, rests, parts, tempo, etc.).

        Returns:
            A new Score instance.

        Examples:
            >>> from aldakit.compose import part, note, tempo
            >>> score = Score.from_elements(
            ...     part("piano"),
            ...     tempo(120),
            ...     note("c", duration=4),
            ...     note("d"),
            ...     note("e"),
            ... )
        """
        score = cls.__new__(cls)
        score._mode = _MODE_ELEMENTS
        score._source = ""
        score._filename = "<compose>"
        score._elements = list(elements)
        score._imported_ast = None
        return score

    @classmethod
    def from_parts(cls, *parts: Any) -> Score:
        """Create a Score from Part objects.

        This is a convenience method for creating a score from parts only.

        Args:
            *parts: Part objects.

        Returns:
            A new Score instance.
        """
        return cls.from_elements(*parts)

    @property
    def source(self) -> str:
        """The original Alda source code (if created from source)."""
        return self._source

    @cached_property
    def ast(self) -> RootNode:
        """The parsed AST (lazily computed and cached)."""
        if self._mode == _MODE_SOURCE:
            return parse(self._source, self._filename)
        elif self._mode == _MODE_MIDI:
            # AST was imported from MIDI file
            assert self._imported_ast is not None
            return self._imported_ast
        else:
            return self._build_ast_from_elements()

    @cached_property
    def midi(self) -> MidiSequence:
        """The generated MIDI sequence (lazily computed and cached)."""
        return generate_midi(self.ast)

    @property
    def duration(self) -> float:
        """Total duration of the score in seconds."""
        return self.midi.duration()

    def _build_ast_from_elements(self) -> RootNode:
        """Build AST directly from compose elements."""
        from .ast_nodes import PartDeclarationNode

        from .compose.part import Part

        children = []
        current_events: list = []
        current_part_decl: PartDeclarationNode | None = None

        def flush_part():
            """Flush accumulated events, wrapping in PartNode if there's a declaration."""
            nonlocal current_events, current_part_decl
            if current_part_decl is not None:
                # Wrap declaration and events in a PartNode
                children.append(
                    PartNode(
                        declaration=current_part_decl,
                        events=EventSequenceNode(events=current_events, position=None),
                        position=None,
                    )
                )
                current_part_decl = None
                current_events = []
            elif current_events:
                # No part declaration - bare event sequence
                children.append(EventSequenceNode(events=current_events, position=None))
                current_events = []

        for element in self._elements:
            ast_node = element.to_ast()

            if isinstance(element, Part):
                # Flush previous part first
                flush_part()
                # Start new part - Part.to_ast() returns PartDeclarationNode
                assert isinstance(ast_node, PartDeclarationNode)
                current_part_decl = ast_node
            else:
                # Accumulate events
                current_events.append(ast_node)

        # Flush remaining content
        flush_part()

        return RootNode(children=children, position=None)

    def _invalidate_cache(self) -> None:
        """Invalidate cached properties after modification."""
        # Delete cached properties if they exist
        for attr in ("ast", "midi"):
            if attr in self.__dict__:
                del self.__dict__[attr]

    # Builder methods

    def add(self, *elements: ComposeElement) -> Score:
        """Add elements to the score.

        This method modifies the score in place and returns self for chaining.
        Only works with scores created via from_elements().

        Args:
            *elements: Compose elements to add.

        Returns:
            Self for method chaining.

        Raises:
            ValueError: If the score was created from source code.
        """
        if self._mode != _MODE_ELEMENTS:
            raise ValueError(
                "Cannot add elements to this score. "
                "Use Score.from_elements() to create a modifiable score."
            )
        self._elements.extend(elements)
        self._invalidate_cache()
        return self

    def with_part(self, *instruments: str, alias: str | None = None) -> Score:
        """Add a part declaration.

        Args:
            *instruments: Instrument names.
            alias: Optional alias for the part.

        Returns:
            Self for method chaining.
        """
        from .compose.part import Part

        return self.add(Part(instruments=instruments, alias=alias))

    def with_tempo(self, bpm: int | float, global_: bool = False) -> Score:
        """Add a tempo attribute.

        Args:
            bpm: Beats per minute.
            global_: If True, applies globally to all parts.

        Returns:
            Self for method chaining.
        """
        from .compose.attributes import Tempo

        return self.add(Tempo(bpm=bpm, global_=global_))

    def with_volume(self, level: int | float) -> Score:
        """Add a volume attribute.

        Args:
            level: Volume level (0-100).

        Returns:
            Self for method chaining.
        """
        from .compose.attributes import Volume

        return self.add(Volume(level=level))

    # Output methods

    def to_alda(self) -> str:
        """Export as Alda source code.

        Returns:
            Alda source code string.
        """
        if self._mode == _MODE_SOURCE:
            return self._source
        elif self._mode == _MODE_MIDI:
            # Generate Alda from AST
            return _ast_to_alda(self.ast)
        else:
            return " ".join(e.to_alda() for e in self._elements)

    def play(
        self,
        port: str | None = None,
        wait: bool = True,
        backend: str = "midi",
        soundfont: str | None = None,
    ) -> None:
        """Play the score.

        Args:
            port: MIDI output port name (for backend="midi"). If None, uses
                the first available port or creates a virtual port.
            wait: If True (default), block until playback completes.
            backend: Backend to use: "midi" (default) or "audio".
                - "midi": Send to MIDI port via libremidi (requires external synth)
                - "audio": Direct audio output via TinySoundFont (requires SoundFont)
            soundfont: Path to SoundFont file (for backend="audio"). If None,
                searches common locations or uses ALDAKIT_SOUNDFONT env var.

        Raises:
            RuntimeError: If the selected backend is not available.
            FileNotFoundError: If backend="audio" and no SoundFont is found.
        """
        if backend == "audio":
            from .midi.backends import TsfBackend, HAS_TSF

            if not HAS_TSF:
                raise RuntimeError(
                    "Audio backend not available. The _tsf module was not built."
                )
            with TsfBackend(soundfont=soundfont) as audio_backend:
                audio_backend.play(self.midi)
                if wait:
                    audio_backend.wait()
        else:
            # Default: MIDI backend
            with LibremidiBackend(port_name=port) as midi_backend:
                midi_backend.play(self.midi)
                if wait:
                    while midi_backend.is_playing():
                        time.sleep(0.1)

    def save(self, path: str | Path) -> None:
        """Save the score to a file.

        Supports both MIDI (.mid, .midi) and Alda (.alda) formats.

        Args:
            path: Output file path.
        """
        path = Path(path)
        if path.suffix in (".mid", ".midi"):
            write_midi_file(self.midi, path)
        elif path.suffix == ".alda":
            path.write_text(self.to_alda(), encoding="utf-8")
        else:
            # Default to MIDI
            write_midi_file(self.midi, path)

    def __repr__(self) -> str:
        if self._mode == _MODE_SOURCE:
            # Show first 50 chars of source, truncated if longer
            preview = self._source[:50]
            if len(self._source) > 50:
                preview += "..."
            preview = preview.replace("\n", "\\n")
            return f"Score({preview!r})"
        elif self._mode == _MODE_MIDI:
            return f"Score.from_midi_file({self._filename!r})"
        else:
            n = len(self._elements)
            return f"Score.from_elements(<{n} elements>)"



from __future__ import annotations

from .async_generator import aclosing, generator_to_async_generator
from .inputhook import (InputHook, InputHookContext, InputHookSelector,
                        new_eventloop_with_inputhook,
                        set_eventloop_with_inputhook)
from .utils import (call_soon_threadsafe, get_traceback_from_context,
                    run_in_executor_with_context)

__all__ = [
    # Async generator
    "generator_to_async_generator",
    "aclosing",
    # Utils.
    "run_in_executor_with_context",
    "call_soon_threadsafe",
    "get_traceback_from_context",
    # Inputhooks.
    "InputHook",
    "new_eventloop_with_inputhook",
    "set_eventloop_with_inputhook",
    "InputHookSelector",
    "InputHookContext",
]



"""Tests for MIDI-level transformers."""

import pytest

from aldakit.midi.types import MidiNote, MidiSequence
from aldakit.midi.transform import (
    # Timing transformers
    quantize,
    humanize,
    swing,
    stretch,
    shift,
    # Velocity transformers
    accent,
    crescendo,
    diminuendo,
    normalize,
    velocity_curve,
    compress,
    # Filtering
    filter_notes,
    trim,
    # Combining
    merge,
    concatenate,
)


def make_sequence(notes: list[tuple[int, int, float, float]]) -> MidiSequence:
    """Helper to create a MidiSequence from note tuples (pitch, velocity, start, duration)."""
    return MidiSequence(
        notes=[
            MidiNote(pitch=p, velocity=v, start_time=s, duration=d)
            for p, v, s, d in notes
        ]
    )


class TestQuantize:
    """Test quantize transformer."""

    def test_quantize_snaps_to_grid(self):
        seq = make_sequence(
            [
                (60, 80, 0.12, 0.5),  # Slightly after 0
                (62, 80, 0.48, 0.5),  # Slightly before 0.5
                (64, 80, 0.76, 0.5),  # Slightly after 0.75
            ]
        )
        quantized = quantize(seq, 0.25)  # Quarter note grid
        assert abs(quantized.notes[0].start_time - 0.0) < 0.01
        assert abs(quantized.notes[1].start_time - 0.5) < 0.01
        assert abs(quantized.notes[2].start_time - 0.75) < 0.01

    def test_quantize_partial_strength(self):
        seq = make_sequence([(60, 80, 0.1, 0.5)])
        # With strength 0.5, should move halfway to grid
        quantized = quantize(seq, 0.25, strength=0.5)
        # 0.1 should move toward 0.0 by 50%: 0.1 - 0.05 = 0.05
        assert abs(quantized.notes[0].start_time - 0.05) < 0.01

    def test_quantize_zero_strength(self):
        seq = make_sequence([(60, 80, 0.1, 0.5)])
        quantized = quantize(seq, 0.25, strength=0.0)
        assert quantized.notes[0].start_time == 0.1

    def test_quantize_preserves_note_properties(self):
        seq = make_sequence([(60, 100, 0.1, 0.5)])
        quantized = quantize(seq, 0.25)
        assert quantized.notes[0].pitch == 60
        assert quantized.notes[0].velocity == 100
        assert quantized.notes[0].duration == 0.5

    def test_quantize_invalid_grid(self):
        seq = make_sequence([(60, 80, 0.1, 0.5)])
        with pytest.raises(ValueError):
            quantize(seq, 0)
        with pytest.raises(ValueError):
            quantize(seq, -0.1)


class TestHumanize:
    """Test humanize transformer."""

    def test_humanize_with_seed_reproducible(self):
        seq = make_sequence([(60, 80, 0.0, 0.5), (62, 80, 0.5, 0.5)])
        h1 = humanize(seq, timing=0.02, velocity=10, seed=42)
        h2 = humanize(seq, timing=0.02, velocity=10, seed=42)
        assert h1.notes[0].start_time == h2.notes[0].start_time
        assert h1.notes[0].velocity == h2.notes[0].velocity

    def test_humanize_timing_within_range(self):
        seq = make_sequence([(60, 80, 1.0, 0.5)])
        humanized = humanize(seq, timing=0.02, seed=123)
        assert 0.98 <= humanized.notes[0].start_time <= 1.02

    def test_humanize_velocity_within_range(self):
        seq = make_sequence([(60, 80, 0.0, 0.5)])
        humanized = humanize(seq, velocity=10, seed=123)
        assert 70 <= humanized.notes[0].velocity <= 90

    def test_humanize_velocity_clamped(self):
        seq = make_sequence([(60, 5, 0.0, 0.5)])  # Low velocity
        humanized = humanize(seq, velocity=20, seed=42)
        assert humanized.notes[0].velocity >= 1

    def test_humanize_duration(self):
        seq = make_sequence([(60, 80, 0.0, 0.5)])
        humanized = humanize(seq, duration=0.05, seed=123)
        assert 0.45 <= humanized.notes[0].duration <= 0.55

    def test_humanize_no_changes_with_zero_amounts(self):
        seq = make_sequence([(60, 80, 0.5, 0.5)])
        humanized = humanize(seq, timing=0.0, velocity=0.0, duration=0.0)
        assert humanized.notes[0].start_time == 0.5
        assert humanized.notes[0].velocity == 80
        assert humanized.notes[0].duration == 0.5


class TestSwing:
    """Test swing transformer."""

    def test_swing_delays_offbeats(self):
        # Notes on grid: 0, 0.5, 1.0, 1.5
        seq = make_sequence(
            [
                (60, 80, 0.0, 0.4),  # On beat
                (62, 80, 0.5, 0.4),  # Offbeat (should be delayed)
                (64, 80, 1.0, 0.4),  # On beat
                (65, 80, 1.5, 0.4),  # Offbeat (should be delayed)
            ]
        )
        swung = swing(seq, amount=0.33, grid=1.0)

        # On-beat notes should not move
        assert swung.notes[0].start_time == 0.0
        assert swung.notes[2].start_time == 1.0

        # Offbeat notes should be delayed
        assert swung.notes[1].start_time > 0.5
        assert swung.notes[3].start_time > 1.5

    def test_swing_amount_affects_delay(self):
        seq = make_sequence([(60, 80, 0.5, 0.4)])  # Offbeat
        light_swing = swing(seq, amount=0.2, grid=1.0)
        heavy_swing = swing(seq, amount=0.5, grid=1.0)

        assert heavy_swing.notes[0].start_time > light_swing.notes[0].start_time

    def test_swing_invalid_grid(self):
        seq = make_sequence([(60, 80, 0.0, 0.5)])
        with pytest.raises(ValueError):
            swing(seq, amount=0.33, grid=0)


class TestStretch:
    """Test stretch transformer."""

    def test_stretch_doubles_time(self):
        seq = make_sequence(
            [
                (60, 80, 0.0, 0.5),
                (62, 80, 1.0, 0.5),
            ]
        )
        stretched = stretch(seq, 2.0)
        assert stretched.notes[0].start_time == 0.0
        assert stretched.notes[0].duration == 1.0
        assert stretched.notes[1].start_time == 2.0
        assert stretched.notes[1].duration == 1.0

    def test_stretch_halves_time(self):
        seq = make_sequence(
            [
                (60, 80, 0.0, 1.0),
                (62, 80, 2.0, 1.0),
            ]
        )
        stretched = stretch(seq, 0.5)
        assert stretched.notes[0].duration == 0.5
        assert stretched.notes[1].start_time == 1.0

    def test_stretch_preserves_pitch_velocity(self):
        seq = make_sequence([(60, 100, 0.0, 0.5)])
        stretched = stretch(seq, 2.0)
        assert stretched.notes[0].pitch == 60
        assert stretched.notes[0].velocity == 100

    def test_stretch_invalid_factor(self):
        seq = make_sequence([(60, 80, 0.0, 0.5)])
        with pytest.raises(ValueError):
            stretch(seq, 0)
        with pytest.raises(ValueError):
            stretch(seq, -1.0)


class TestShift:
    """Test shift transformer."""

    def test_shift_positive(self):
        seq = make_sequence(
            [
                (60, 80, 0.0, 0.5),
                (62, 80, 1.0, 0.5),
            ]
        )
        shifted = shift(seq, 1.0)
        assert shifted.notes[0].start_time == 1.0
        assert shifted.notes[1].start_time == 2.0

    def test_shift_negative_removes_notes(self):
        seq = make_sequence(
            [
                (60, 80, 0.0, 0.5),
                (62, 80, 1.0, 0.5),
            ]
        )
        shifted = shift(seq, -0.6)
        # First note starts at -0.6, ends at -0.1, so it's completely before 0
        # Second note starts at 0.4
        assert len(shifted.notes) == 1
        assert abs(shifted.notes[0].start_time - 0.4) < 0.01

    def test_shift_negative_truncates_note(self):
        seq = make_sequence([(60, 80, 0.0, 1.0)])
        shifted = shift(seq, -0.3)
        # Note would start at -0.3 but extends to 0.7, so truncate
        assert shifted.notes[0].start_time == 0.0
        assert abs(shifted.notes[0].duration - 0.7) < 0.01


class TestAccent:
    """Test accent transformer."""

    def test_accent_pattern(self):
        seq = make_sequence(
            [
                (60, 80, 0.0, 0.5),
                (62, 80, 0.5, 0.5),
                (64, 80, 1.0, 0.5),
                (65, 80, 1.5, 0.5),
            ]
        )
        accented = accent(seq, [1.25, 0.75])
        assert accented.notes[0].velocity == 100  # 80 * 1.25
        assert accented.notes[1].velocity == 60  # 80 * 0.75
        assert accented.notes[2].velocity == 100  # Pattern repeats
        assert accented.notes[3].velocity == 60

    def test_accent_with_base_velocity(self):
        seq = make_sequence([(60, 80, 0.0, 0.5)])
        accented = accent(seq, [1.0], base_velocity=100)
        assert accented.notes[0].velocity == 100

    def test_accent_empty_pattern(self):
        seq = make_sequence([(60, 80, 0.0, 0.5)])
        accented = accent(seq, [])
        assert accented.notes[0].velocity == 80  # Unchanged

    def test_accent_clamped(self):
        seq = make_sequence([(60, 120, 0.0, 0.5)])
        accented = accent(seq, [2.0])  # Would be 240
        assert accented.notes[0].velocity == 127


class TestCrescendo:
    """Test crescendo transformer."""

    def test_crescendo_basic(self):
        seq = make_sequence(
            [
                (60, 80, 0.0, 0.5),
                (62, 80, 1.0, 0.5),
                (64, 80, 2.0, 0.5),
            ]
        )
        cresc = crescendo(seq, start_velocity=40, end_velocity=100)
        assert cresc.notes[0].velocity == 40
        assert cresc.notes[1].velocity == 70  # Midpoint
        assert cresc.notes[2].velocity == 100

    def test_crescendo_time_range(self):
        seq = make_sequence(
            [
                (60, 80, 0.0, 0.5),
                (62, 80, 1.0, 0.5),
                (64, 80, 2.0, 0.5),
                (65, 80, 3.0, 0.5),
            ]
        )
        cresc = crescendo(seq, 40, 100, start_time=1.0, end_time=2.0)
        assert cresc.notes[0].velocity == 40  # Before range
        assert cresc.notes[1].velocity == 40  # Start of range
        assert cresc.notes[2].velocity == 100  # End of range
        assert cresc.notes[3].velocity == 100  # After range

    def test_crescendo_empty_sequence(self):
        seq = MidiSequence()
        cresc = crescendo(seq, 40, 100)
        assert len(cresc.notes) == 0


class TestDiminuendo:
    """Test diminuendo transformer."""

    def test_diminuendo_basic(self):
        seq = make_sequence(
            [
                (60, 80, 0.0, 0.5),
                (62, 80, 1.0, 0.5),
                (64, 80, 2.0, 0.5),
            ]
        )
        dim = diminuendo(seq, start_velocity=100, end_velocity=40)
        assert dim.notes[0].velocity == 100
        assert dim.notes[2].velocity == 40


class TestNormalize:
    """Test normalize transformer."""

    def test_normalize_scales_to_target(self):
        seq = make_sequence(
            [
                (60, 50, 0.0, 0.5),
                (62, 100, 0.5, 0.5),  # Max velocity
                (64, 75, 1.0, 0.5),
            ]
        )
        normalized = normalize(seq, target=80)
        # Scale factor: 80/100 = 0.8
        assert normalized.notes[0].velocity == 40  # 50 * 0.8
        assert normalized.notes[1].velocity == 80  # 100 * 0.8
        assert normalized.notes[2].velocity == 60  # 75 * 0.8

    def test_normalize_empty_sequence(self):
        seq = MidiSequence()
        normalized = normalize(seq, target=100)
        assert len(normalized.notes) == 0


class TestVelocityCurve:
    """Test velocity_curve transformer."""

    def test_velocity_curve_custom(self):
        seq = make_sequence([(60, 80, 0.0, 0.5)])
        curved = velocity_curve(seq, lambda v: v // 2)
        assert curved.notes[0].velocity == 40

    def test_velocity_curve_clamped(self):
        seq = make_sequence([(60, 80, 0.0, 0.5)])
        curved = velocity_curve(seq, lambda v: v * 3)  # Would be 240
        assert curved.notes[0].velocity == 127


class TestCompress:
    """Test compress transformer."""

    def test_compress_above_threshold(self):
        seq = make_sequence(
            [
                (60, 60, 0.0, 0.5),  # Below threshold
                (62, 100, 0.5, 0.5),  # Above threshold
            ]
        )
        compressed = compress(seq, threshold=80, ratio=2.0)
        assert compressed.notes[0].velocity == 60  # Unchanged
        # 100 is 20 above threshold, compressed: 80 + 20/2 = 90
        assert compressed.notes[1].velocity == 90

    def test_compress_invalid_ratio(self):
        seq = make_sequence([(60, 80, 0.0, 0.5)])
        with pytest.raises(ValueError):
            compress(seq, threshold=80, ratio=0)


class TestFilterNotes:
    """Test filter_notes transformer."""

    def test_filter_by_velocity(self):
        seq = make_sequence(
            [
                (60, 50, 0.0, 0.5),
                (62, 100, 0.5, 0.5),
                (64, 30, 1.0, 0.5),
            ]
        )
        filtered = filter_notes(seq, lambda n: n.velocity > 40)
        assert len(filtered.notes) == 2
        assert filtered.notes[0].velocity == 50
        assert filtered.notes[1].velocity == 100

    def test_filter_by_pitch(self):
        seq = make_sequence(
            [
                (48, 80, 0.0, 0.5),  # Low
                (60, 80, 0.5, 0.5),  # Middle
                (72, 80, 1.0, 0.5),  # High
            ]
        )
        # Keep only middle register (C4 to C5)
        filtered = filter_notes(seq, lambda n: 60 <= n.pitch <= 72)
        assert len(filtered.notes) == 2

    def test_filter_by_time(self):
        seq = make_sequence(
            [
                (60, 80, 0.0, 0.5),
                (62, 80, 1.0, 0.5),
                (64, 80, 2.0, 0.5),
            ]
        )
        filtered = filter_notes(seq, lambda n: n.start_time >= 1.0)
        assert len(filtered.notes) == 2


class TestTrim:
    """Test trim transformer."""

    def test_trim_basic(self):
        seq = make_sequence(
            [
                (60, 80, 0.0, 0.5),
                (62, 80, 1.0, 0.5),
                (64, 80, 2.0, 0.5),
                (65, 80, 3.0, 0.5),
            ]
        )
        trimmed = trim(seq, start=1.0, end=3.0)
        assert len(trimmed.notes) == 2
        # Times should be shifted to start at 0
        assert trimmed.notes[0].start_time == 0.0
        assert trimmed.notes[1].start_time == 1.0

    def test_trim_preserves_note_properties(self):
        seq = make_sequence([(60, 100, 1.0, 0.5)])
        trimmed = trim(seq, start=0.0, end=2.0)
        assert trimmed.notes[0].pitch == 60
        assert trimmed.notes[0].velocity == 100
        assert trimmed.notes[0].duration == 0.5


class TestMerge:
    """Test merge transformer."""

    def test_merge_two_sequences(self):
        seq1 = make_sequence([(60, 80, 0.0, 1.0)])
        seq2 = make_sequence([(64, 80, 0.0, 1.0)])
        merged = merge(seq1, seq2)
        assert len(merged.notes) == 2

    def test_merge_sorts_by_time(self):
        seq1 = make_sequence([(60, 80, 1.0, 0.5)])
        seq2 = make_sequence([(64, 80, 0.0, 0.5)])
        merged = merge(seq1, seq2)
        assert merged.notes[0].start_time == 0.0
        assert merged.notes[1].start_time == 1.0

    def test_merge_empty(self):
        merged = merge()
        assert len(merged.notes) == 0


class TestConcatenate:
    """Test concatenate transformer."""

    def test_concatenate_two_sequences(self):
        seq1 = make_sequence(
            [
                (60, 80, 0.0, 0.5),
                (62, 80, 0.5, 0.5),
            ]
        )  # Duration: 1.0
        seq2 = make_sequence(
            [
                (64, 80, 0.0, 0.5),
            ]
        )
        concat = concatenate(seq1, seq2)
        assert len(concat.notes) == 3
        assert concat.notes[0].start_time == 0.0
        assert concat.notes[1].start_time == 0.5
        assert concat.notes[2].start_time == 1.0  # After seq1 ends

    def test_concatenate_with_gap(self):
        seq1 = make_sequence([(60, 80, 0.0, 0.5)])  # Duration: 0.5
        seq2 = make_sequence([(64, 80, 0.0, 0.5)])
        concat = concatenate(seq1, seq2, gap=0.5)
        # seq1 ends at 0.5, gap of 0.5, seq2 starts at 1.0
        assert concat.notes[1].start_time == 1.0

    def test_concatenate_empty(self):
        concat = concatenate()
        assert len(concat.notes) == 0


class TestIntegration:
    """Integration tests for MIDI transformers."""

    def test_humanize_then_quantize(self):
        """Humanize and then re-quantize."""
        seq = make_sequence(
            [
                (60, 80, 0.0, 0.5),
                (62, 80, 0.5, 0.5),
                (64, 80, 1.0, 0.5),
            ]
        )
        humanized = humanize(seq, timing=0.1, seed=42)
        requantized = quantize(humanized, 0.5)

        # After re-quantizing, notes should be close to original grid
        assert abs(requantized.notes[0].start_time - 0.0) < 0.1
        assert abs(requantized.notes[1].start_time - 0.5) < 0.1
        assert abs(requantized.notes[2].start_time - 1.0) < 0.1

    def test_stretch_and_crescendo(self):
        """Stretch timing and apply crescendo."""
        seq = make_sequence(
            [
                (60, 80, 0.0, 0.5),
                (62, 80, 0.5, 0.5),
            ]
        )
        stretched = stretch(seq, 2.0)
        cresc = crescendo(stretched, 40, 100)

        assert stretched.notes[1].start_time == 1.0
        assert cresc.notes[0].velocity == 40
        assert cresc.notes[1].velocity == 100

    def test_with_score(self):
        """Test using MIDI transforms with a Score."""
        from aldakit import Score
        from aldakit.compose import part, note, tempo

        score = Score.from_elements(
            part("piano"),
            tempo(120),
            note("c", duration=8),
            note("d", duration=8),
            note("e", duration=8),
        )

        midi = score.midi
        humanized = humanize(midi, timing=0.01, velocity=5, seed=42)

        assert len(humanized.notes) == 3
        # Notes should be slightly different from original
        # (can't assert exact values due to randomness, but structure preserved)



"""Token types and Token class for the Alda lexer."""

from dataclasses import dataclass
from enum import Enum, auto
from typing import Any


class TokenType(Enum):
    """All token types recognized by the Alda scanner."""

    # Note and rest letters
    NOTE_LETTER = auto()  # a-g
    REST_LETTER = auto()  # r

    # Accidentals
    FLAT = auto()  # -
    SHARP = auto()  # +
    NATURAL = auto()  # _

    # Octave control
    OCTAVE_SET = auto()  # o followed by number
    OCTAVE_UP = auto()  # >
    OCTAVE_DOWN = auto()  # <

    # Duration
    NOTE_LENGTH = auto()  # numeric duration (1, 2, 4, 8, 16, etc.)
    NOTE_LENGTH_MS = auto()  # duration in milliseconds (e.g., 500ms)
    NOTE_LENGTH_SECONDS = auto()  # duration in seconds (e.g., 2s)
    DOT = auto()  # . for dotted notes

    # Connectors
    TIE = auto()  # ~
    BARLINE = auto()  # |
    SEPARATOR = auto()  # / for chords

    # Part declaration
    NAME = auto()  # identifier (instrument name or variable)
    ALIAS = auto()  # quoted string for part alias
    COLON = auto()  # :

    # S-expression tokens (lisp mode)
    LEFT_PAREN = auto()  # (
    RIGHT_PAREN = auto()  # )
    SYMBOL = auto()  # lisp symbol
    NUMBER = auto()  # numeric literal in lisp context
    STRING = auto()  # string literal in lisp context

    # Variables
    EQUALS = auto()  # = for variable definition

    # Markers
    MARKER = auto()  # %name
    AT_MARKER = auto()  # @name

    # Voice markers
    VOICE_MARKER = auto()  # V[number]:

    # Cram expressions
    CRAM_OPEN = auto()  # {
    CRAM_CLOSE = auto()  # }

    # Event sequences
    EVENT_SEQ_OPEN = auto()  # [
    EVENT_SEQ_CLOSE = auto()  # ]

    # Repetition
    REPEAT = auto()  # *[number]
    REPETITIONS = auto()  # '[ranges] for on-repetition (e.g., '1-3,5)

    # Control
    EOF = auto()
    NEWLINE = auto()


@dataclass(frozen=True)
class SourcePosition:
    """Position in source code for error reporting."""

    line: int
    column: int
    filename: str = "<input>"

    def __str__(self) -> str:
        return f"{self.filename}:{self.line}:{self.column}"


@dataclass
class Token:
    """A token produced by the scanner."""

    type: TokenType
    lexeme: str
    literal: Any  # Parsed value (e.g., int for NOTE_LENGTH)
    position: SourcePosition

    def __repr__(self) -> str:
        if self.literal is not None and self.literal != self.lexeme:
            return f"Token({self.type.name}, {self.lexeme!r}, {self.literal!r})"
        return f"Token({self.type.name}, {self.lexeme!r})"
